// $Id: action.h,v 1.98 2004/09/21 15:13:09 uckelman Exp $
/* Board -- a graphical frontend for boardgames
 * Copyright (C) 2003, 2004 Joel Uckelman
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __ACTION_H_
#define __ACTION_H_

#include <qcanvas.h>
#include <qdom.h>
#include <qpixmap.h>


class Group;
class Highlight;
class Player;
class Polyline;
class Surface;
class Token;
class TokenDrag;
class View;

struct TokenType;


class Action {
   public:
      Action();
      Action(unsigned int o);
      virtual ~Action();
      virtual bool commit();
      virtual bool undo();
      virtual void begin();
      virtual void end();
      virtual void abort();
      virtual void preUndo();
      virtual bool isPermissible() const = 0;
      virtual bool canPrecede(const Action &a) const;
      virtual QPixmap icon() const = 0;
      virtual QString name() const = 0;
      void setUpdate(bool u);
      virtual QValueList<QCanvas *> canvasList() const;
      virtual QCanvasItemList createdList() const;
      Player *owner() const;
      virtual QString text() const;
      virtual QPoint center() const;
      virtual unsigned int stepForward();
      virtual unsigned int stepBackward();
      virtual unsigned int currentStep() const;
      virtual unsigned int commitStep() const;
      virtual unsigned int undoStep() const;
      virtual unsigned int firstStep() const;
      virtual unsigned int lastStep() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const = 0;
      
   protected:
      // for commitImp, undoImp, focus, unfocus, highlight, unhighlight
      friend class ComboAction; 

      virtual void focus();
      virtual void unfocus();
      virtual void highlight();
      virtual void unhighlight();
      virtual bool commitImp();
      virtual bool undoImp();

      QDomElement prepareXML(QDomElement e) const;

      unsigned int oid;
      int step; 
      bool update;
};


class ComboAction : public Action {
   public:
      ComboAction();
      ComboAction(unsigned int o);
      virtual ~ComboAction();
      virtual unsigned int stepForward();
      virtual unsigned int stepBackward();
      virtual bool commit();
      virtual bool undo();
      virtual void end();
      virtual void abort();
      virtual void preUndo();
      virtual bool isPermissible() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual QValueList<QCanvas *> canvasList() const;
      virtual QCanvasItemList createdList() const;
      virtual void append(Action *a);
      virtual void prepend(Action *a);
      unsigned int count() const;
      virtual Action *collapse();
      virtual QString text() const;
      virtual QPoint center() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      virtual void focus();
      QDomElement prepareXML(QDomElement e) const;

      QValueList<Action *> list;

   private:
      void update();
};


class TokenAction : public Action {
   public:
      TokenAction(Token *t);
      TokenAction(unsigned int o, unsigned int t);
      virtual ~TokenAction();
      virtual QValueList<QCanvas *> canvasList() const;
      virtual QPoint center() const;

   protected:
      virtual void highlight();
      virtual void unhighlight();
      virtual void focus();

      QDomElement prepareXML(QDomElement e) const;

      unsigned int tid;
      Highlight *hlight;
};


class TokenFlipAction : public TokenAction {
   public:
      TokenFlipAction(Token *t, bool d);
      TokenFlipAction(unsigned int o, unsigned int t, bool d);
      virtual bool isPermissible() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      bool next();
      bool previous();
      QDomElement prepareXML(QDomElement e) const;

   private:
      bool direction;
};


class TokenFlipLeftAction : public TokenFlipAction {
   public:
      TokenFlipLeftAction(Token *t);
      TokenFlipLeftAction(unsigned int o, unsigned int t);
      virtual QPixmap icon() const;
      virtual QString name() const;

   private:
      virtual bool commitImp();
      virtual bool undoImp();

      static const QPixmap *Icon;
};


class TokenFlipRightAction : public TokenFlipAction {
   public:
      TokenFlipRightAction(Token *t);
      TokenFlipRightAction(unsigned int o, unsigned int t);
      virtual QPixmap icon() const;
      virtual QString name() const;

   private:
      virtual bool commitImp();
      virtual bool undoImp();

      static const QPixmap *Icon;
};


class TokenRotateAction : public TokenAction {
   public:
      TokenRotateAction(Token *t, int dt);
      TokenRotateAction(unsigned int o, unsigned int t, int dt);
      virtual bool isPermissible() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      virtual bool commitImp();
      virtual bool undoImp();
      QDomElement prepareXML(QDomElement e) const;

   private:
      int dtheta;
};


class TokenRotateCWAction : public TokenRotateAction {
   public:
      TokenRotateCWAction(Token *t);
      TokenRotateCWAction(unsigned int o, unsigned int t, int dt);
      virtual QPixmap icon() const;
      virtual QString name() const;

   private:
      static const QPixmap *Icon;
};


class TokenRotateCCWAction : public TokenRotateAction {
   public:
      TokenRotateCCWAction(Token *t);
      TokenRotateCCWAction(unsigned int o, unsigned int t, int dt);
      virtual QPixmap icon() const;
      virtual QString name() const;

   private:
      static const QPixmap *Icon;
};


class TokenMoveAction : public TokenAction {
   public:
      TokenMoveAction(Token *t, double xp, double yp, double zp);
      TokenMoveAction(Token *t, QCanvas *dc, double xp, double yp, double zp);
      TokenMoveAction(Token *t, QCanvas *dc);
      TokenMoveAction(Token *t);
      TokenMoveAction(unsigned int o, unsigned int t, unsigned int s,
       double xp, double yp, double zp);
      virtual ~TokenMoveAction();
      virtual bool isPermissible() const;
      virtual QValueList<QCanvas *> canvasList() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual void begin();
      virtual void end();
      virtual QString text() const;
      virtual QPoint center() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      virtual void highlight();
      virtual void unhighlight();
      virtual bool commitImp();
      virtual bool undoImp();
      QDomElement prepareXML(QDomElement e) const;

      double x, y, z;
      unsigned int sid;

      Highlight *here, *there;
      QCanvasLine *line;

   private:
      static const QPixmap *Icon;
};


class TokenCreateAction: public TokenAction {
   public:
      TokenCreateAction(TokenType *d, Group *g);
      TokenCreateAction(unsigned int o, unsigned int t);
      Token *created() const;
      virtual void abort();
      virtual void preUndo();
      virtual bool isPermissible() const;
      virtual QValueList<QCanvas *> canvasList() const;
      virtual QCanvasItemList createdList() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual void begin();
      virtual void end();
      virtual QPoint center() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      virtual void highlight();
      virtual void focus();
      virtual bool commitImp();
      virtual bool undoImp();
      TokenCreateAction(Token *c, Group *g);

      unsigned int gid, sid, type;
      double x, y, z;
      QString Name;

      static const QPixmap *Icon;
};


class TokenCloneAction : public TokenCreateAction {
   public:
      TokenCloneAction(Token *c, Group *g);
      TokenCloneAction(Token *c, Group *g, QCanvas *dc, double xp, double yp,
       double zp);
      TokenCloneAction(unsigned int o, unsigned int t, unsigned int c);
      virtual ~TokenCloneAction();
      virtual void preUndo();
      virtual void begin();
      virtual void abort();
      virtual bool isPermissible() const;
      virtual QValueList<QCanvas *> canvasList() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual QPoint center() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      virtual void highlight();
      virtual void unhighlight();
      virtual void focus();
      virtual bool commitImp();
      virtual bool undoImp();
      QDomElement prepareXML(QDomElement e) const;

   private:
      unsigned int copyof, up;
      int theta;

      Highlight *here, *there;
      QCanvasLine *line;

      static const QPixmap *Icon;
};


class TokenTrashAction : public TokenAction {
   public:
      TokenTrashAction(Token *t);
      TokenTrashAction(unsigned int o, unsigned int t, unsigned int s,
       unsigned int g, double ox, double oy, double oz, int a,
      const QString &n, unsigned int tp, unsigned int u);
      virtual bool isPermissible() const;
      virtual QValueList<QCanvas *> canvasList() const;
      virtual QPoint center() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      virtual void highlight();
      virtual void focus();
      virtual bool commitImp();
      virtual bool undoImp();
      QDomElement prepareXML(QDomElement e) const;

   private:
      unsigned int sid, gid, type, up;
      double x, y, z;
      int theta;
      QString Name;

      static const QPixmap *Icon;
};


class TokenRegroupAction : public TokenAction {
   public:
      TokenRegroupAction(Token *t);
      TokenRegroupAction(Token *t, Group *p);
      TokenRegroupAction(unsigned int o, unsigned int t, unsigned int p);
      virtual bool isPermissible() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      void setParent(Group *p);
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      virtual bool commitImp();
      virtual bool undoImp();
      QDomElement prepareXML(QDomElement e) const;

   private:
      unsigned int gid;

      static const QPixmap *Icon;
};


class TokenRenameAction : public TokenAction {
   public:
      TokenRenameAction(Token *t, const QString &n);
      TokenRenameAction(unsigned int o, unsigned int t, const QString &n);
      virtual bool isPermissible() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      virtual bool commitImp();
      virtual bool undoImp();
      QDomElement prepareXML(QDomElement e) const;

   private:
      QString Name;

      static const QPixmap *Icon;
};


class TokenStackAction : public ComboAction {
   public:
      TokenStackAction();
      TokenStackAction(unsigned int o); 
      virtual QString name() const;
      virtual QPixmap icon() const;
      virtual Action *collapse();
      virtual QDomElement dumpXML(QDomDocument &doc) const;
      
   private:
      static const QPixmap *Icon;
};


class TokenInvertAction : public TokenStackAction {
   public:
      TokenInvertAction();
      TokenInvertAction(unsigned int o);
      virtual QString name() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;
};


class TokenShuffleAction : public TokenStackAction {
   public:
      TokenShuffleAction();
      TokenShuffleAction(unsigned int o);
      virtual QString name() const;
      virtual QPixmap icon() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   private:
      static const QPixmap *Icon;
};


class PolylineAction : public Action {
   public:
      PolylineAction(Polyline *l);
      PolylineAction(unsigned int o, unsigned int l);
  
   protected:
      QDomElement prepareXML(QDomElement e) const;

      unsigned int lid;
};


class PolylineBeginAction : public PolylineAction {
   public:
      PolylineBeginAction(Group *g, QCanvas *c, double xp, double yp,
       const QString &n, QPen p);
      PolylineBeginAction(unsigned int o, unsigned int g, unsigned int s,
       unsigned int l, double xp, double yp, const QString &n, QPen p); 
      Polyline *created() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual bool isPermissible() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      QDomElement prepareXML(QDomElement e) const;

   private:
      virtual bool commitImp();
      virtual bool undoImp();

      unsigned int gid, sid;
      double x, y;
      QString Name;
      QPen pen;

      static const QPixmap *Icon;
};


class PolylineExtendAction : public PolylineAction {
   public:
      PolylineExtendAction(Polyline *l, double xp, double yp, QPen p);
      PolylineExtendAction(unsigned int o, unsigned int l, double xp,
       double yp, QPen p); 
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual bool isPermissible() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      QDomElement prepareXML(QDomElement e) const;

   private:
      virtual bool commitImp();
      virtual bool undoImp();

      double x, y;
      QPen pen;

      static const QPixmap *Icon;
};


class GroupAction : public Action {
   public:
      GroupAction(Group *g);
      GroupAction(unsigned int o, unsigned int g);
   
   protected:
      QDomElement prepareXML(QDomElement e) const;

      unsigned int gid;
};


class GroupCreateAction : public GroupAction {
   public:
      GroupCreateAction(const QString &n, Group *p);
      GroupCreateAction(unsigned int o, unsigned int g);
      Group *created() const;
      virtual bool isPermissible() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   private:
      virtual bool commitImp(); 
      virtual bool undoImp();

      unsigned int pid;
      QString Name;

      static const QPixmap *Icon;
};


class GroupTrashAction : public ComboAction {
   public:
      GroupTrashAction(Group *g);
      GroupTrashAction(unsigned int o, unsigned int g, unsigned int p,
       const QString &n);
      virtual bool isPermissible() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      virtual bool commitImp();
      virtual bool undoImp();
      QDomElement prepareXML(QDomElement e) const;

   private:
      unsigned int gid, pid; 
      QString Name;

      static const QPixmap *Icon;
};


class GroupRegroupAction : public GroupAction {
   public:
      GroupRegroupAction(Group *g);
      GroupRegroupAction(unsigned int o, unsigned int g, unsigned int p);
      virtual bool isPermissible() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual void begin();
      void setParent(Group *p);
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      virtual bool commitImp();
      virtual bool undoImp();
      QDomElement prepareXML(QDomElement e) const;

   private:
      unsigned int pid;

      static const QPixmap *Icon;
};


class GroupRenameAction : public GroupAction {
   public:
      GroupRenameAction(Group *g, const QString &n);
      GroupRenameAction(unsigned int o, unsigned int g, const QString &n);
      virtual bool isPermissible() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      virtual bool commitImp();
      virtual bool undoImp();
      QDomElement prepareXML(QDomElement e) const;

      QString Name;
      
      static const QPixmap *Icon;
};


class GroupSetPermissionAction : public GroupAction {
   public:
      GroupSetPermissionAction(Group *g, Player *p, unsigned int pm, int v);
      GroupSetPermissionAction(unsigned int o, unsigned int g, unsigned int p,
       unsigned int pm, int v);
      virtual bool isPermissible() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      virtual bool commitImp();
      virtual bool undoImp();
      QDomElement prepareXML(QDomElement e) const;

   private:
      unsigned int pid;
      unsigned int perm;
      int val;

      static const QPixmap *Icon;
};


class GroupVisibilityAction : public GroupAction {
   public:
      GroupVisibilityAction(Group *g);
      GroupVisibilityAction(unsigned int o, unsigned int g);

   protected:
      bool show();
      bool hide();

   private:
      void showUpwards(Group *g) const;
      void hideUpwards(Group *g) const;
};


class GroupShowAction : public GroupVisibilityAction {
   public:
      GroupShowAction(Group *g);
      GroupShowAction(unsigned int o, unsigned int g);
      virtual bool isPermissible() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   private:
      virtual bool commitImp();
      virtual bool undoImp();

      static const QPixmap *Icon;
};


class GroupHideAction : public GroupVisibilityAction {
   public:
      GroupHideAction(Group *g);
      GroupHideAction(unsigned int o, unsigned int g);
      virtual bool isPermissible() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   private:
      virtual bool commitImp();
      virtual bool undoImp();

      static const QPixmap *Icon;
};


class PlayerAction : public Action {
   public:
      PlayerAction(Player *p);
      PlayerAction(unsigned int o, unsigned int p);

   protected:
      QDomElement prepareXML(QDomElement e) const;

      unsigned int pid;
};


class PlayerCreateAction : public PlayerAction {
   public:
      PlayerCreateAction(const QString &n);
      PlayerCreateAction(unsigned int o, unsigned int p);
      Player *created() const;
      virtual bool isPermissible() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   private:
      virtual bool commitImp();
      virtual bool undoImp();

      QString Name;

      static const QPixmap *Icon;
};


class PlayerTrashAction : public ComboAction {
   public:
      PlayerTrashAction(Player *p);
      PlayerTrashAction(unsigned int o, unsigned int p, const QString &n,
       const QString &pw);
      virtual bool isPermissible() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      virtual bool commitImp();
      virtual bool undoImp();
      QDomElement prepareXML(QDomElement e) const;

   private:
      unsigned int pid;
      QString Name, Pass;

      static const QPixmap *Icon;
};


class PlayerRenameAction : public GroupRenameAction {
   public:
      PlayerRenameAction(Player *p, const QString &n);
      PlayerRenameAction(unsigned int o, unsigned int p, const QString &n);
      virtual QString name() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      QDomElement prepareXML(QDomElement e) const;

   private:
      static const QPixmap *Icon;
};


class SurfaceAction : public Action {
   public:
      SurfaceAction(Surface *s);
      SurfaceAction(unsigned int o, unsigned int s);

   protected:
      QDomElement prepareXML(QDomElement e) const;

      unsigned int sid;
};


class SurfaceCreateAction : public SurfaceAction {
   public:
      SurfaceCreateAction(const QString &n);
      SurfaceCreateAction(unsigned int o, unsigned int s);
      Surface *created() const;
      virtual bool isPermissible() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   private:
      virtual bool commitImp();
      virtual bool undoImp();

      QString Name;

      static const QPixmap *Icon;
};


class SurfaceTrashAction : public ComboAction {
   public:
      SurfaceTrashAction(Surface *s);
      SurfaceTrashAction(unsigned int o, unsigned int s, const QString &n,
       const QString &b);
      virtual bool isPermissible() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      virtual bool commitImp();
      virtual bool undoImp();
      QDomElement prepareXML(QDomElement e) const;

   private:
      unsigned int sid;
      QString Name, background;

      static const QPixmap *Icon;
};


class SurfaceRenameAction : public SurfaceAction {
   public:
      SurfaceRenameAction(Surface *s, const QString &n);
      SurfaceRenameAction(unsigned int o, unsigned int s, const QString &n);
      virtual bool isPermissible() const;
      virtual QPixmap icon() const;
      virtual QString name() const;
      virtual QDomElement dumpXML(QDomDocument &doc) const;

   protected:
      virtual bool commitImp();
      virtual bool undoImp();
      QDomElement prepareXML(QDomElement e) const;

   private:
      QString Name;

      static const QPixmap *Icon;
};

#endif /* __ACTION_H_ */
