// $Id: editview.cpp,v 1.12 2004/04/02 23:24:49 uckelman Exp $
/* Board -- a graphical frontend for boardgames
 * Copyright (C) 2003, 2004 Joel Uckelman
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <qapplication.h>
#include <qpixmap.h>
#include <qvbox.h>

#include "editview.h"


const int BoxSize = 16;


//
// RenameEdit
//
RenameEdit::RenameEdit(QWidget *parent) : QLineEdit(parent) {
}


void RenameEdit::keyPressEvent(QKeyEvent *e) {
   if (e->key() == Key_Escape) emit escapePressed();
   else QLineEdit::keyPressEvent(e);
   e->accept();
}


void RenameEdit::focusOutEvent(QFocusEvent *) { emit escapePressed(); }


//
// EditView
//
EditView::EditView(QWidget *parent, const char *name) :
 QListView(parent, name), renameItem(0), renaming(false),
 firstMousePressEvent(true) {
   QVBox *vb = new QVBox(viewport());
   vb->setFrameStyle(QFrame::Box | QFrame::Plain);
   edit = new RenameEdit(vb);
   edit->setFixedHeight(edit->sizeHint().height());
   vb->hide();
   vb->setBackgroundMode(PaletteBase);
   edit_timer = new QTimer(this);
   connect(edit, SIGNAL(returnPressed()),
    this, SLOT(rename()));
   connect(edit, SIGNAL(escapePressed()),
    this, SLOT(cancelRename()));
   connect(edit_timer, SIGNAL(timeout()),
    this, SLOT(doubleClickTimeout()));
}


void EditView::keyPressEvent(QKeyEvent *e) {
   if ((e->key() == Key_Enter || e->key() == Key_Return) && renaming) return;
   cancelRename();
   QListView::keyPressEvent(e);
}


void EditView::viewportMousePressEvent(QMouseEvent *e) {
   bool didRename = renaming;
   cancelRename();
   
   if (!hasFocus() && !viewport()->hasFocus()) setFocus();

   if (e->button() != LeftButton) {
      QListView::viewportMousePressEvent(e);
      firstMousePressEvent = false;
      return;
   }

   QListViewItem *i = currentItem();
   QListView::viewportMousePressEvent(e);

   if (itemAt(e->pos()) != i || e->x() + contentsX() > columnWidth(0)) {
      firstMousePressEvent = false;
      return;
   }

   if (!firstMousePressEvent && !didRename && i == currentItem() && i) {
      edit_timer->start(QApplication::doubleClickInterval(), true);
      renameItem = i;
   }

   firstMousePressEvent = false;
}


void EditView::viewportMouseDoubleClickEvent(QMouseEvent *e) {
   edit_timer->stop();
   QListView::viewportMouseDoubleClickEvent(e);
}


void EditView::viewportMouseMoveEvent(QMouseEvent *e) {
   edit_timer->stop();
   QListView::viewportMouseMoveEvent(e);
}


void EditView::doubleClickTimeout() {
   startRename();
   edit_timer->stop();
}


void EditView::startRename(bool check) {
   if (check && (!renameItem || renameItem != currentItem())) return;

   QListViewItem *i = currentItem();
   setSelected(i, true);

   QRect r = itemRect(i);
   
   int offset = 
    treeStepSize()*(i->depth() + (rootIsDecorated() ? 1 : 0)) +
    (hasCheckBox.find(i) != hasCheckBox.end() ?
     itemMargin() + BoxSize + itemMargin() : 0) +
    (i->pixmap(0) ? i->pixmap(0)->width() + itemMargin() : 0);

   int x = r.x() + offset;
   int y = r.y();
   int w = columnWidth(0) - offset;
   int h = QMAX(edit->height() + 2, r.height());
   y += r.height()/2 - h/2;
   
   edit->parentWidget()->setGeometry(x, y, w+6, h);
   edit->setFocus();
   edit->setText(i->text(0));
   edit->selectAll();
   edit->setFrame(false);
   edit->parentWidget()->show();
   viewport()->setFocusProxy(edit);
   renaming = true;
}


void EditView::clear() {
   cancelRename();
   QListView::clear();
}


void EditView::rename() {
   QString et = edit->text();
   if (!et.isEmpty()) {
      QString n = currentItem()->text(0);
      if (et != n) currentItem()->setText(0, et);
   }
   cancelRename();
}


void EditView::cancelRename() {
   renameItem = 0;
   edit->parentWidget()->hide();
   viewport()->setFocusProxy(this);
   renaming = false;
   if (currentItem()) currentItem()->repaint();
   if (edit->hasFocus()) viewport()->setFocus();
}

