// $Id: group.cpp,v 1.45 2004/08/12 22:16:37 uckelman Exp $
/* Board -- a graphical frontend for boardgames
 * Copyright (C) 2003, 2004 Joel Uckelman
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <typeinfo>

#include "game.h"
#include "group.h"
#include "player.h"
#include "token.h"


//
// Perms
//
int Perms::get(const Player *player, unsigned int perm) const {
   if (p2p.contains(player) && p2p[player].contains(perm))
    return p2p[player][perm];
   else return -1;
}


QMap<unsigned int, bool> Perms::get(const Player *player) const {
   if (p2p.contains(player)) return p2p[player];
   else return QMap<unsigned int, bool>();
}


void Perms::set(const Player *player, unsigned int perm, bool val) {
   if (!p2p.contains(player)) p2p[player] = QMap<unsigned int, bool>();
   p2p[player][perm] = val;
}


void Perms::set(const Player *player, const QString &s) {
   if (!p2p.contains(player)) p2p[player] = QMap<unsigned int, bool>();

   for (unsigned int i = 0; i < PNum; i++) {
      switch(s.at(i)) {
         case '0': p2p[player][i] = false; break;
         case '1': p2p[player][i] = true;  break;
         case '-': p2p[player].remove(i);  break;
      }
   }
}


void Perms::unset(const Player *player, unsigned int perm) {
   if (!p2p.contains(player)) return;
   p2p[player].remove(perm);
}


QString Perms::toString(const Player *player) const {
   QString s;   
   if (!p2p.contains(player)) {
      s.fill('-', 10);
      return s;
   }

   for (unsigned int i = 0; i < PNum; i++) {
      if (p2p[player].contains(i))
       s += p2p[player][i] ? '1' : '0';
      else s += '-';
   }
   return s;
}


bool Perms::contains(const Player *p) const { return p2p.contains(p); }


//
// Groupable
//
Groupable::Groupable(GameObj *g, const QString &n, Group *p, unsigned int i) :
 game(g), Name(n), Parent(p), ID(i ? i : game->maxID++) {
   if (i >= game->maxID) game->maxID = i + 1;
   if (Parent) Parent->append(this);
}


Groupable::~Groupable() {
   if (Parent) Parent->remove(this);
}


QString Groupable::name() const { return Name; }


void Groupable::setName(const QString &n) { Name = n; }


QString Groupable::notes() const { return Notes; }


void Groupable::setNotes(const QString &n) { Notes = n; }


Player *Groupable::owner() {
   if (Parent) return Parent->owner();
   else return 0;
}


Group *Groupable::parent() const { return Parent; }


void Groupable::reparent(Group *p) {
   if (Parent) Parent->remove(this);
   if (p) p->append(this);
   Parent = p;
}


unsigned int Groupable::id() const { return ID; }


//
// Group
//
Group::Group(GameObj *g, const QString &n, Group *gr, unsigned int i) :
 Groupable(g, n, gr, i), Selected(false) {
   Visible[game->user] = true;
}


Group::~Group() {
   while (!isEmpty()) delete first();
}


bool Group::permission(const Player *pl, unsigned int pm) const {
   int ret;
   if ((ret = perms.get(pl, pm)) != -1) return ret;
   else if (Parent) return Parent->permission(pl, pm);
   else return true;
}


int Group::localPermission(const Player *pl, unsigned int pm) const {
   return perms.get(pl, pm);   
}


bool Group::inheritedPermission(const Player *pl, unsigned int pm) const {
   return Parent ? Parent->permission(pl, pm) : true;
}


void Group::setPermission(const Player *pl, unsigned int pm, bool val) {
   perms.set(pl, pm, val);
}


void Group::unsetPermission(const Player *pl, unsigned int pm) {
   perms.unset(pl, pm);
}

void Group::setPermissions(const Player *pl, const QString &s) {
   perms.set(pl, s);
}


QDomElement Group::dumpXML(QDomDocument &doc) const {
   QDomElement e = doc.createElement("group");
   e.setAttribute("name", Name);
   e.setAttribute("id", id());

   // state
   QValueList<Player *> plist = game->root->players();
   for (QValueList<Player *>::const_iterator i = plist.begin();
    i != plist.end(); ++i) {
      QDomElement s = doc.createElement("state");
      s.setAttribute("pid", (*i)->id());

      if (perms.contains(*i))
       s.setAttribute("permissions", perms.toString(*i));
 
      e.appendChild(s);
   }

   // children
   for (const_iterator i = begin(); i != end(); ++i)
    e.appendChild((*i)->dumpXML(doc));
   return e;
}


bool Group::selected() const { return Selected; }


void Group::setSelected(bool s) { Selected = s; }


bool Group::isVisible() const { return Visible[game->user]; }


void Group::setVisibility(bool v) { Visible[game->user] = v; }


void Group::setVisibility(Player *p, bool v) { Visible[p] = v; }


void Group::show() {
   /*
   if (permission(game->user, Perms::Show)) {
      Visible[game->user] = true;
      for (const_iterator i = begin(); i != end(); ++i) (*i)->show();
   }
   */

   for (const_iterator i = begin(); i != end(); ++i)
    if (dynamic_cast<Token *>(*i)) (*i)->show();

   Visible[game->user] = true;
}


void Group::hide() {
   /*
   if (permission(game->user, Perms::Hide)) {
      Visible[game->user] = false;
      for (const_iterator i = begin(); i != end(); ++i) (*i)->hide();
   }
   */
   for (const_iterator i = begin(); i != end(); ++i)
    if (dynamic_cast<Token *>(*i)) (*i)->hide();

   Visible[game->user] = false;
}


QValueList<Token *> Group::tokens() const {
   QValueList<Token *> list;
   for (const_iterator i = begin(); i != end(); ++i) {
      if (typeid(*i) == typeid(Token *)) list.append((Token *)*i);
      else {
         Group *g = dynamic_cast<Group *>(*i);
         if (g) list += g->tokens();
      }
   }
   return list;
}


QValueList<QCanvas *> Group::canvases() const {
   QValueList<QCanvas *> list;
   Token *t;
   Group *g;

   for (const_iterator i = begin(); i != end(); ++i) {
      if ((t = dynamic_cast<Token *>(*i))) {
         if (list.find(t->canvas()) == list.end()) list.append(t->canvas());
      }
      else if ((g = dynamic_cast<Group *>(*i))) {
         QValueList<QCanvas *> rec = g->canvases();
         for (QValueList<QCanvas *>::const_iterator r = rec.begin();
          r != rec.end(); ++r)
          if (list.find(*r) == list.end()) list.append(*r);
      }
   }
   return list;
}


//
// RootGroup
//
RootGroup::RootGroup(GameObj *g, unsigned int i) : Group(g, "All", 0, i) {
}


QDomElement RootGroup::dumpXML(QDomDocument &doc) const {
   QDomElement e = doc.createElement("entities");
   e.setAttribute("id", id());

   // children
   for (const_iterator i = begin(); i != end(); ++i)
    e.appendChild((*i)->dumpXML(doc));
   return e;
}


QValueList<Player *> RootGroup::players() const {
   QValueList<Player *> list;
   for (const_iterator i = begin(); i != end(); ++i) {
      Player *p = dynamic_cast<Player *>(*i);
      if (p) list.append(p);
   }
   return list;
}
