// $Id: groupdialogs.cpp,v 1.67 2004/08/12 22:06:03 uckelman Exp $
/* Board -- a graphical frontend for boardgames
 * Copyright (C) 2002, 2003, 2004 Joel Uckelman
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <qapplication.h>
#include <qdatastream.h>
#include <qheader.h>
#include <qinputdialog.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qpainter.h>
#include <qpushbutton.h>
#include <qtooltip.h>

#include "action.h"
#include "bijection.h"
#include "game.h"
#include "groupdialogs.h"
#include "main.h"
#include "player.h"
#include "playerprop.h"

#include "tools/clone.xpm"
#include "tools/eye.xpm"
#include "tools/flipright.xpm"
#include "tools/info.xpm"
#include "tools/insert.xpm"
#include "tools/move.xpm"
#include "tools/new.xpm"
#include "tools/prop.xpm"
#include "tools/remove.xpm"
#include "tools/rotateccw.xpm"
#include "tools/trash.xpm"

extern GameObj *Game;
extern QString GroupTreeItemMIMEtype;
extern Main *MW;

const int BoxSize = 16;

// FIXME: rename groups via double click on group name
// FIXME: multiple selections

//
// GroupTreeView
//
GroupTreeView::GroupTreeView(GameObj *g, QWidget *parent, const char *name) :
 EditView(parent, name), pressed(false), oldCurrent(0), dropItem(0), game(g) {
   addColumn("Name");
   header()->hide();
   setRootIsDecorated(true);

   autoopen_timer = new QTimer(this);
   setAcceptDrops(true);
   viewport()->setAcceptDrops(true);
   connect(autoopen_timer, SIGNAL(timeout()), this, SLOT(openGroup()));
}


void GroupTreeView::setSelected(QListViewItem *i, bool select) {
   QListView::setSelected(i, select);
   if (i == firstChild()) emit rootSelected(select);
}


void GroupTreeView::contentsDragEnterEvent(QDragEnterEvent *e) {
   if (!GroupTreeItemDrag::canDecode(e)) {
      e->ignore();
      return;
   }

   oldCurrent = currentItem();
   QListViewItem *i = itemAt(contentsToViewport(e->pos()));
   if (i) {
      dropItem = i;
      autoopen_timer->start(autoopen_time);
   }
}


void GroupTreeView::contentsDragMoveEvent(QDragMoveEvent *e) {
   if (!GroupTreeItemDrag::canDecode(e)) {
      e->ignore();
      return;
   }

   QPoint vp = contentsToViewport(e->pos());
   QListViewItem *i = itemAt(vp);
   if (i) {
      setSelected(i, true);
      e->accept();
      if (i != dropItem) {
         autoopen_timer->stop();
         dropItem = i;
         autoopen_timer->start(autoopen_time);
      }
      switch (e->action()) {
         case QDropEvent::Move: 
         case QDropEvent::Link:
            e->acceptAction();
            break;
         case QDropEvent::Copy:
         default:
            break;
      }
   }
   else {
      e->ignore();
      autoopen_timer->stop();
      dropItem = 0;
   }
}


void GroupTreeView::contentsDragLeaveEvent(QDragLeaveEvent *) {
   autoopen_timer->stop();
   dropItem = 0;
   setCurrentItem(oldCurrent);
   setSelected(oldCurrent, true);
}


void GroupTreeView::contentsDropEvent(QDropEvent *e) {
   autoopen_timer->stop();

   if (!GroupTreeItemDrag::canDecode(e)) {
      e->ignore();
      return;
   }

   QListViewItem *dest = itemAt(contentsToViewport(e->pos()));
   if (dest) {
      QListViewItem *rg; 
      if (GroupTreeItemDrag::decode(e, rg)) {
         // FIXME: what about actions other than move?
         switch (e->action()) {
            case QDropEvent::Move:
            case QDropEvent::Link:
               e->acceptAction();
               break;
            case QDropEvent::Copy:
            default:
               break;
         }
   
         e->accept();

         // get RegroupAction
         Action *a = game->done.top();

         // set destination Group
         GroupTreeDialog *dialog = (GroupTreeDialog *)parent();

         Group *gd = dynamic_cast<Group *>(dialog->g2i[dest]);
         if (!gd) gd = dialog->g2i[dest]->parent();

         GroupRegroupAction *gra;
         TokenRegroupAction *tra;
         if ((gra = dynamic_cast<GroupRegroupAction *>(a))) gra->setParent(gd);
         else if ((tra = dynamic_cast<TokenRegroupAction *>(a)))
          tra->setParent(gd);   
          
         // commit, or abort if destination is impermissible
         if (a->isPermissible()) {
            a->commit();
            game->done.complete();
         }
         else game->done.abort();
      }
   }
   else e->ignore();
}


void GroupTreeView::contentsMouseMoveEvent(QMouseEvent *e) {
   // start drag only if cursor has moved far enough since button press
   if (pressed && (presspos - e->pos()).manhattanLength() >
    QApplication::startDragDistance()) {
      pressed = false;
      
      QListViewItem *i = itemAt(contentsToViewport(presspos));
      if (!i) return;
         
      Groupable *gr = ((GroupTreeDialog *)parent())->g2i[i];

      // do not allow Players or Root to be dragged
      if (gr == game->root || dynamic_cast<Player *>(gr)) return;
      
      Action *a = 0;
      Token *c;
      Group *g;

      if ((c = dynamic_cast<Token *>(gr))) a = new TokenRegroupAction(c);
      else if ((g = dynamic_cast<Group *>(gr))) a = new GroupRegroupAction(g); 
         
      if (a && a->isPermissible()) {
         a->begin();
         game->done.log(a);
 
         GroupTreeItemDrag *gd = new GroupTreeItemDrag(i, viewport());
         gd->drag();
      }
      else delete a;
   }
}


void GroupTreeView::contentsMousePressEvent(QMouseEvent *e) {
   // set presspos here so GroupCheckItem::activate() can use it when
   // called from QListView::contentsMousePressEvent()
   presspos = e->pos();

   QListView::contentsMousePressEvent(e);

   // check whether press was over an item
   QPoint p(contentsToViewport(presspos));
   QListViewItem *i = itemAt(p);
   if (!i) return;
 
   int boxLeft = treeStepSize()*(i->depth()+1);

   // record press if it wasn't over the check box
   if (p.x() > boxLeft + itemMargin() + BoxSize + 4 || p.x() < boxLeft)
    pressed = true;
}


void GroupTreeView::contentsMouseReleaseEvent(QMouseEvent *) {
   pressed = false;
}


void GroupTreeView::rename() {
   QString et = edit->text();
   if (!et.isEmpty()) {
      QString n = renameItem->text(0);
      if (et != n) {
         Groupable *g = ((GroupTreeDialog *)parent())->g2i[renameItem];
         Action *a = 0;
        
         if (dynamic_cast<Token *>(g))
          a = new TokenRenameAction((Token *)g, et);
         else if (dynamic_cast<Player *>(g))
          a = new PlayerRenameAction((Player *)g, et);
         else if (dynamic_cast<Group *>(g))
          a = new GroupRenameAction((Group *)g, et);
          
         if (a && a->isPermissible()) {
            a->commit();
            game->done.log(a);
            game->done.complete();
         }
         else delete a;
      }
   }
   cancelRename();
}


void GroupTreeView::openGroup() {
   autoopen_timer->stop();
   if (dropItem && !dropItem->isOpen()) {
      dropItem->setOpen(true);
      dropItem->repaint();
   }
}


//
// GroupTreeCheckItem
//
GroupTreeCheckItem::GroupTreeCheckItem(GroupTreeView *parent,
 const QString &text) : QCheckListItem(parent, text, CheckBox) {
   if (parent) parent->hasCheckBox.append(this);
}


GroupTreeCheckItem::GroupTreeCheckItem(QListViewItem *parent,
 const QString &text) : QCheckListItem(parent, text, CheckBox) {
   if (listView()) ((GroupTreeView *)listView())->hasCheckBox.append(this);
}


GroupTreeCheckItem::~GroupTreeCheckItem() {
   if (listView()) ((GroupTreeView *)listView())->hasCheckBox.remove(this);
}


void GroupTreeCheckItem::constructVisibilityAction(Group *g,
 ComboAction *combo) const {

   Action *a;
   if (isOn()) a = new GroupHideAction(g);
   else a = new GroupShowAction(g);

   if (a->isPermissible()) {   
      combo->append(a);

      for (Group::const_iterator i = g->begin(); i != g->end(); ++i)
       if (dynamic_cast<Group *>(*i))
        constructVisibilityAction((Group *)*i, combo);
   }
   else delete a;
}


void GroupTreeCheckItem::activate() {
   GroupTreeView *view = (GroupTreeView *)listView();

   QPoint p(view->contentsToViewport(view->presspos));
   int boxLeft = view->treeStepSize()*(depth()+1);

   // do nothing if click wasn't over the check box
   if (p.x() > boxLeft + view->itemMargin() + BoxSize + 4 || p.x() < boxLeft)
    return;

   // otherwise prepare a show or hide action
   Group *g =
    (Group *)(((GroupTreeDialog *)(listView()->parent()))->g2i[this]);
   ComboAction *combo = new ComboAction;
   constructVisibilityAction(g, combo);

   Action *a = combo->collapse();
   if (!a) return;
   
   QCheckListItem::activate();
   
   a->commit();
   Game->done.log(a);
   Game->done.complete();
}


//
// GroupTreeItemDrag
//
GroupTreeItemDrag::GroupTreeItemDrag(QListViewItem *i, QWidget *dragSource,
 const char *name) : QStoredDrag(GroupTreeItemMIMEtype, dragSource, name) {
   QByteArray bytes;
   QDataStream str(bytes, IO_WriteOnly);
   str << (Q_ULONG)i;
   setEncodedData(bytes);
}


bool GroupTreeItemDrag::canDecode(const QMimeSource *m) {
   return m->provides(GroupTreeItemMIMEtype);
}


bool GroupTreeItemDrag::decode(const QMimeSource *m, QListViewItem *&i) {
   QDataStream str(m->encodedData(GroupTreeItemMIMEtype), IO_ReadOnly);
   Q_ULONG ptr;
   str >> ptr;
   i = (QListViewItem *)ptr;
   return true;
}


//
// GroupTreeDialog
//
GroupTreeDialog::GroupTreeDialog(GameObj *g, QWidget *parent, const char *name,
 WFlags f) : TabPage(QIconSet(), "Groups", parent, name, f), game(g) {
   QVBoxLayout *vl = (QVBoxLayout *)layout();

   // view
   view = new GroupTreeView(game, this);
   connect(view, SIGNAL(doubleClicked(QListViewItem *)),
    this, SLOT(showPermissions(QListViewItem *)));
   connect(view, SIGNAL(selectionChanged(QListViewItem *)),
    this, SLOT(changedSelection(QListViewItem *)));
   connect(view,
    SIGNAL(contextMenuRequested(QListViewItem *, const QPoint &, int)),
    this, SLOT(showContextMenu(QListViewItem *, const QPoint &)));
   vl->addWidget(view);
   
   // buttons
   QHBoxLayout *bl = new QHBoxLayout(vl);

   // new 
   nb = new QPushButton(this);
   nb->setPixmap(QPixmap((const char **)new_xpm));
   nb->setFixedSize(27, 27);
   QToolTip::add(nb, "New");
   connect(nb, SIGNAL(clicked()), this, SLOT(createGroup()));
   bl->addWidget(nb);

   // trash
   tb = new QPushButton(this);
   tb->setPixmap(QPixmap((const char **)trash_xpm));
   tb->setFixedSize(27, 27);
   QToolTip::add(tb, "Delete");
   connect(tb, SIGNAL(clicked()), this, SLOT(deleteGroup()));
   bl->addWidget(tb);

   bl->insertStretch(-1);
}


Group *GroupTreeDialog::currentGroup() const {
   QListViewItem *cur = view->selectedItem();
   if (cur) {
      Token *c = dynamic_cast<Token *>(g2i[cur]);
      if (c) return c->parent();
      else return dynamic_cast<Group *>(g2i[cur]);
   }
   else return game->root;
}


void GroupTreeDialog::setVisibility(Group *g, bool v) {
   if (!g2i.contains(g)) return;
   QCheckListItem *item = (QCheckListItem *)g2i[g];
   
   if (v && !(item->isOn())) item->setOn(true);
   else if (!v && item->isOn()) item->setOn(false);
}


void GroupTreeDialog::insert(Groupable *g) {
   QListViewItem *i;
   if (dynamic_cast<Token *>(g)) {
      if (g->parent()) i = new QListViewItem(g2i[g->parent()], g->name());
      else i = new QListViewItem(view, g->name());
      g2i.insert(g, i);
   }
   else {
      if (g->parent()) i = new GroupTreeCheckItem(g2i[g->parent()], g->name());
      else i = new GroupTreeCheckItem(view, g->name());
      g2i.insert(g, i);
      ((GroupTreeCheckItem *)i)->setOn(((Group *)g)->isVisible());
   }
}


void GroupTreeDialog::remove(Groupable *g) {
   Group *gp = dynamic_cast<Group *>(g);
   if (gp) {
      for (Group::const_iterator i = gp->begin(); i != gp->end(); ++i)
       remove(*i);
   }

   delete g2i[g];
   g2i.remove(g);
}


void GroupTreeDialog::reparent(Groupable *g) {
   QListViewItem *src = g2i[g];
   QListViewItem *dest = g2i[g->parent()];

   src->parent()->takeItem(src);
   dest->insertItem(src);
}


void GroupTreeDialog::rename(Groupable *g) {
   QListViewItem *i = g2i[g];
   i->setText(0, g->name());
}


void GroupTreeDialog::createGroup() {
   // get group name
   bool ok = false;
   QString name = QInputDialog::getText("Create a Group", "Group Name:",
    QLineEdit::Normal, QString::null, &ok, this);   
   if (!ok) return;

   // take action
   GroupCreateAction *a = new GroupCreateAction(name, currentGroup());
   if (a->isPermissible()) {
      a->commit();
      game->done.log(a);
      game->done.complete();
   }
   else delete a;
}


void GroupTreeDialog::deleteGroup() {
   QListViewItem *i = view->selectedItem();
   if (!i || !g2i.contains(i)) return;

   Groupable *gr = g2i[i];
   if (gr == game->root) return;
   
   Action *a = 0;
   Token *c = dynamic_cast<Token *>(gr);
   if (c) a = new TokenTrashAction(c);
   else {
     Group *g = dynamic_cast<Group *>(gr); 
     if (g) a = new GroupTrashAction(g);
   }

   if (a && a->isPermissible()) {
      a->commit();
      game->done.log(a);
      game->done.complete();
   }
   else delete a;   
}


void GroupTreeDialog::showPermissions() {
   showPermissions(view->currentItem());
}


void GroupTreeDialog::showPermissions(QListViewItem *i) {
   Group *g = dynamic_cast<Group *>(g2i[i]);
   if (g) {
      if (g->permission(game->user, Perms::Properties))
       GroupPropertiesDialog::editPerms(g, this);
      else GroupPropertiesDialog::showPerms(g, this);    
   }
}


void GroupTreeDialog::changedSelection(QListViewItem *i) {
   // FIXME: buttons should be updated when permissions change as well

   // disable new if user lacks permission to create under this group
   Groupable *g = g2i[i];
   nb->setEnabled(g && g->permission(game->user, Perms::Insert));

   // disable trash if no selection or Root is selected
   tb->setEnabled((g && g != game->root)); 
}


void GroupTreeDialog::showProperties() {
   Groupable *g = g2i[view->currentItem()];

   if (dynamic_cast<Player *>(g)) {
      if (PlayerPropertiesDialog::setProperties(game, (Player *)g, this))
       MW->setChanged(true);
   }
}


void GroupTreeDialog::showContextMenu(QListViewItem *i, const QPoint &p) {
   if (!i) return;

   QPopupMenu menu(this);
   menu.insertItem("Properties...", this, SLOT(showProperties()));
   menu.insertItem("Permissions...", this, SLOT(showPermissions()));
   menu.exec(p);
}


//
// GroupPermListItem
//
GroupPermListItem::GroupPermListItem(QListView *parent, const QString &name) :
 QListViewItem(parent, name), enabled(true) {
}


int GroupPermListItem::width(const QFontMetrics &fm, const QListView *lv,
 int column) const {
   return QListViewItem::width(fm, lv, column);
}


void GroupPermListItem::paintCell(QPainter *p, const QColorGroup &cg,
 int column, int width, int align) {
   if (!p) return;
   if (column == 0) {
      QListViewItem::paintCell(p, cg, column, width, align);
      return;
   }

   p->fillRect(0, 0, width, height(), cg.brush(QColorGroup::Base));

   QListView *lv = listView();
   if (!lv) return;

   unsigned int j = listView()->header()->mapToSection(column)-1;
   int x = (width - BoxSize) / 2;
   int y = (height() - BoxSize) / 2;
  
   // FIXME: do something to indicate items are disabled for read-only
   if (State[j] == 2) p->setPen(QPen(lv->palette().color(QPalette::Disabled,
    QColorGroup::Text), 2));
   else p->setPen(QPen(cg.text(), 2));
 
   p->drawRect(x+2, y+2, BoxSize-4, BoxSize-4);
   
   if (State[j] == 1 || (State[j] == 2 && Inherited[j])) {
      x += 3 + lv->itemMargin();
      y += 6;

      QPointArray a(7*2);
      for (int i = 0; i < 3; i++, x++, y++) {
         a.setPoint(2*i, x, y);
         a.setPoint(2*i+1, x, y+2);
      }
      
      y -= 2;
      for (int i = 3; i < 7; i++, x++, y--) {
         a.setPoint(2*i, x, y);
         a.setPoint(2*i+1, x, y+2);
      }
      p->drawLineSegments(a);
   }
}


void GroupPermListItem::setState(int column, unsigned int state) {
   if (state == State[column-1]) return;
   State[column-1] = state;
   repaint();
} 

unsigned int GroupPermListItem::state(int column) const {
   return State[column-1];
}


void GroupPermListItem::setInherited(int column, bool in) {
   if (in == Inherited[column-1]) return;
   Inherited[column-1] = in;
   if (State[column-1] == 2) repaint();
} 


bool GroupPermListItem::inherited(int column) const {
   return Inherited[column-1];
}


void GroupPermListItem::activate() {
   if ((listView() && !listView()->isEnabled()) || !isEnabled()) return;

   QPoint pos;
   activatedPos(pos);

   QHeader *h = listView()->header();

   // ignore clicks on the name column
   int column = h->sectionAt(pos.x());
   if (column == 0) return;

   // ignore clicks outside of check boxes
   int x = h->sectionPos(column);
   if (pos.x() < 4 + x || pos.x() > 4 + x + BoxSize) return;

   // cycle: off -> on -> inherited
   setState(column, (state(column)+1)%3);
}


bool GroupPermListItem::isEnabled() const { return enabled; }


void GroupPermListItem::setEnabled(bool e) { enabled = e; }


//
// GroupPropertiesDialog
//
const QPixmap *GroupPropertiesDialog::Icon[] = { 
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };


GroupPropertiesDialog::GroupPropertiesDialog(Group *g, bool edit,
 QWidget *parent, const char *name, bool modal, WFlags f) :
 QDialog(parent, name, modal, f), group(g) {
   setCaption(QString(edit ? "Edit" : "View") + " Group Properties");

   QVBoxLayout *vb = new QVBoxLayout(this, 5);

   view = new QListView(this);
   vb->addWidget(view);
  
   view->addColumn("Player");
   view->addColumn(QIconSet(QPixmap((const char **)eye_xpm)), "", 26); 
   // FIXME: make an X'd eye for hide
   view->addColumn(QIconSet(QPixmap((const char **)eye_xpm)), "", 26); 
   view->addColumn(QIconSet(QPixmap((const char **)move_xpm)), "", 26); 
   view->addColumn(QIconSet(QPixmap((const char **)clone_xpm)), "", 26); 
   view->addColumn(QIconSet(QPixmap((const char **)rotateccw_xpm)), "", 26); 
   view->addColumn(QIconSet(QPixmap((const char **)flipright_xpm)), "", 26); 
   view->addColumn(QIconSet(QPixmap((const char **)trash_xpm)), "", 26); 
   view->addColumn(QIconSet(QPixmap((const char **)insert_xpm)), "", 26); 
   view->addColumn(QIconSet(QPixmap((const char **)remove_xpm)), "", 26); 
   view->addColumn(QIconSet(QPixmap((const char **)info_xpm)), "", 26); 
   view->addColumn(QIconSet(QPixmap((const char **)prop_xpm)), "", 26); 

   view->header()->setResizeEnabled(false);
   view->header()->setMovingEnabled(false);
   view->setSelectionMode(QListView::NoSelection);

   QValueList<Player *> plist = Game->root->players();
   for (QValueList<Player *>::const_iterator i = plist.begin();
    i != plist.end(); ++i) {
      if (*i != group->owner()) {
         GroupPermListItem *lvi = new GroupPermListItem(view, (*i)->name());
         i2p.insert(lvi, *i);
         if (!edit) lvi->setEnabled(false);
        
         for (unsigned int j = 0; j < Perms::PNum; j++) {
            lvi->setInherited(j+1, group->inheritedPermission(*i, j));
            int local = group->localPermission(*i, j);
            lvi->setState(j+1, local == -1 ? 2 : local);
         }
      }
   }
  
   // buttons
   QHBoxLayout *hb = new QHBoxLayout(vb);
   hb->addStretch();

   QPushButton *ok = new QPushButton("&Ok", this);
   connect(ok, SIGNAL(clicked()), this, SLOT(accept()));
   ok->setDefault(true);
   hb->addWidget(ok);

   if (edit) {
      QPushButton *cancel = new QPushButton("&Cancel", this);
      connect(cancel, SIGNAL(clicked()), this, SLOT(reject()));
      hb->addWidget(cancel);
   }
}


void GroupPropertiesDialog::editPerms(Group *g, QWidget *parent) {
   GroupPropertiesDialog pd(g, true, parent);
   if ((pd.exec() == Accepted)) {
      // apply changes
      ComboAction *combo = new ComboAction;
      for (QListViewItemIterator i(pd.view); i.current() != 0; ++i) {
         Player *player = pd.i2p[i.current()];
         for (unsigned int j = 0; j < Perms::PNum; j++) {
            int state = ((GroupPermListItem *)(i.current()))->state(j+1);
            if (state == 2) state = -1;
            int prev = pd.group->localPermission(player, j);

            if (prev != state) {
               Action *a = new
                GroupSetPermissionAction(pd.group, player, j, state);
               if (a->isPermissible()) combo->append(a);
               else delete a;   
            }
         }
      }

      Action *a = combo->collapse();
      if (a) {
         a->commit();
         Game->done.log(a);
         Game->done.complete();
      }
   }
}


void GroupPropertiesDialog::showPerms(Group *g, QWidget *parent) {
   GroupPropertiesDialog pd(g, false, parent);
   pd.exec();
}

/*
//
// GroupCreateDialog
//
GroupCreateDialog::GroupCreateDialog(QWidget *parent, const char *name,
 bool modal, WFlags f) : QDialog(parent, name, modal, f) {
   setCaption("Create a Group");
   
   QVBoxLayout *vb = new QVBoxLayout(this, 5);

   // first row
   QHBoxLayout *hb = new QHBoxLayout(vb);

   // combo box
   combo = new QComboBox(false, this);
   combo->insertItem("Group");
   combo->insertItem("Player");
   
   // name edit   
   ne = new QLineEdit(this);
   ne->setFocus();

   hb->addWidget(combo);
   hb->addWidget(ne);

   // second row
   hb = new QHBoxLayout(vb);
   hb->addStretch();
   
   // buttons
   QPushButton *ok = new QPushButton("&Ok", this);
   connect(ok, SIGNAL(clicked()), this, SLOT(accept()));
   ok->setDefault(true);
   hb->addWidget(ok);

   QPushButton *cancel = new QPushButton("&Cancel", this);
   connect(cancel, SIGNAL(clicked()), this, SLOT(reject()));
   hb->addWidget(cancel);   
}


QString GroupCreateDialog::getName(QWidget *parent) {
   QString name;
   GroupCreateDialog cd(parent);
   if ((cd.exec() == Accepted)) {
      name = cd.ne->text();
   }
   return name;
}
*/
