// $Id: history.cpp,v 1.13 2004/08/26 20:47:48 uckelman Exp $
/* Board -- a graphical frontend for boardgames
 * Copyright (C) 2004 Joel Uckelman
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <qlayout.h>
#include <qlistbox.h>
#include <qpushbutton.h>
#include <qtimer.h>
#include <qtooltip.h>

#include "game.h"
#include "history.h"
#include "selectionhandler.h"

#include "icons/first.xpm"
#include "icons/last.xpm"
#include "icons/next.xpm"
#include "icons/play.xpm"
#include "icons/prev.xpm"
#include "icons/rewind.xpm"
#include "icons/stop.xpm"


//
// HistoryListItem
//
HistoryListItem::HistoryListItem(const QString &text, Action *a) :
 QListBoxText(text), actionPtr(a) {
}


Action *HistoryListItem::action() const { return actionPtr; }


//
// HistoryListBox
//
HistoryListBox::HistoryListBox(GameObj *g, QWidget *parent, const char *name,
 WFlags f) : QListBox(parent, name, f), game(g) {
}


void HistoryListBox::finalize() {
   // make an initial element
   insertItem(new HistoryListItem("Initial"), 0);

   // make a list item for each Action
   QListBoxItem *lbi;

   for (LogStack::const_iterator i = game->done.begin(); i != game->done.end();
    ++i) insertItem(lbi = new HistoryListItem((*i)->name(), *i));

   // make an element separating history from the undo stack
   //insertItem(new HistoryListItem(QString::null), -1);

   // make last item current
   setCurrentItem(count()-1);
}


void HistoryListBox::mousePressEvent(QMouseEvent *e) {
   int i = index(itemAt(e->pos()));

   if (i == -1) e->ignore();
   else {
      if (timer) stopTimer();
      i > currentItem() ? skipForward(i) : skipBackward(i) ;
      e->accept();
   }
}


void HistoryListBox::keyPressEvent(QKeyEvent *e) {
   if (count() < 2) {
      e->ignore();
      return;
   }

   switch(e->key()) {
      case Key_Up:    if (currentItem() > 0) prev();                break;
      case Key_Prior: if (currentItem() > 0) pgUp();                break;
      case Key_Home:  if (currentItem() > 0) first();               break;
      case Key_Down:  if (currentItem() < (int)count() - 1) next(); break;
      case Key_Next:  if (currentItem() < (int)count() - 1) pgDn(); break;
      case Key_End:   if (currentItem() < (int)count() - 1) last(); break;
   }

   e->accept();
}


void HistoryListBox::first() {
   if (timer) stopTimer();
   skipBackward(0);
}


void HistoryListBox::last() {
   if (timer) stopTimer();
   skipForward(count()-1);
}


void HistoryListBox::pgUp() {
   if (timer) stopTimer();

   int i = currentItem() - numItemsVisible();
   i = i < 0 ? 0 : i;
   skipBackward(i);
   setTopItem(i);
}


void HistoryListBox::pgDn() {
   if (timer) stopTimer();
   
   int i = currentItem() + numItemsVisible();
   i = i > (int)count() - 1 ? (int)count() - 1 : i;
   skipForward(i);
   setBottomItem(i);
}


void HistoryListBox::prev() {
   // clear the current selection
   game->selector->clearSelection();

   // start the timer
   if (timer) stopTimer();
   timer = new QTimer(this);
   connect(timer, SIGNAL(timeout()), this, SLOT(stepBackwardOne()));
   timer->start(150);
}


void HistoryListBox::next() {
   // clear the current selection
   game->selector->clearSelection();

   // start the timer
   if (timer) stopTimer();
   timer = new QTimer(this);
   connect(timer, SIGNAL(timeout()), this, SLOT(stepForwardOne()));
   timer->start(150);
}


void HistoryListBox::rewind() {
   // clear the current selection
   game->selector->clearSelection();

   // start the timer
   if (timer) stopTimer();
   timer = new QTimer(this);
   connect(timer, SIGNAL(timeout()), this, SLOT(stepBackwardContinuous()));
   timer->start(250);
}


void HistoryListBox::play() {
   // clear the current selection
   game->selector->clearSelection();

   // start the timer
   if (timer) stopTimer();
   timer = new QTimer(this);
   connect(timer, SIGNAL(timeout()), this, SLOT(stepForwardContinuous()));
   timer->start(250);
}


void HistoryListBox::stop() {
   // kill the timer
   stopTimer();

   // finish the current Action
   Action *a = ((HistoryListItem *)item(currentItem()))->action();
   if (a) {
      if (a->currentStep() > a->commitStep()) 
       while (a->stepBackward() != a->commitStep());
      else if (a->currentStep() < a->commitStep())
       while (a->stepForward() != a->commitStep());
   }
}


void HistoryListBox::stopTimer() {
   delete timer;
   timer = 0;
}


bool HistoryListBox::skipBackward(int target) {
   if (target < 0 || target >= (int)count()) return false;

   // clear the current selection
   game->selector->clearSelection();

   int current = currentItem();
   Action *a;

   // skip backwards
   while (target < current) {
      if ((a = ((HistoryListItem *)item(current))->action())) {
         a->preUndo();
         if (a->isPermissible()) {
            a = game->done.pop();
            a->undo();
            game->undone.push(a);
            --current;
         }      
         else return false;
      }
      else --current;      
   }

   setCurrentItem(target);
   return true;
}


bool HistoryListBox::skipForward(int target) {
   if (target < 0 || target >= (int)count()) return false;

   // clear the current selection
   game->selector->clearSelection();

   int current = currentItem();
   Action *a;

   // skip forwards
   while (target > current) {
      if ((a = ((HistoryListItem *)item(current+1))->action())) {
         if (a->isPermissible()) {
            a = game->undone.pop();
            a->commit();
            game->done.push(a);
            ++current;
         }
         else return false;
      }
      else ++current; 
   }

   setCurrentItem(target);
   return true;
}


void HistoryListBox::stepBackwardContinuous() {
   Action *a = ((HistoryListItem *)item(currentItem()))->action();
   if (a) {
      a->preUndo();
      if (a->isPermissible()) {
         if (a->stepBackward() == a->firstStep()) {
            game->undone.push(game->done.pop());
            if (currentItem() > 0) setCurrentItem(currentItem()-1);
            else stopTimer();
         }
      }
      else stopTimer();
   }
   else if (currentItem() > 0) setCurrentItem(currentItem()-1);
   else stopTimer();
}


void HistoryListBox::stepForwardContinuous() {
   Action *a = ((HistoryListItem *)item(currentItem()))->action();
   if (a) {
      if (a->isPermissible()) {
         if (a->stepForward() == a->lastStep()) {
            game->done.push(game->undone.pop());
            if (currentItem() < (int)count() - 1)
             setCurrentItem(currentItem()+1);
            else stopTimer();
         }
      }
      else stopTimer();
   }
   else if (currentItem() < (int)count() - 1) setCurrentItem(currentItem()+1);
   else stopTimer();
}


void HistoryListBox::stepBackwardOne() {
   static int advanced = 0;

   Action *curr = ((HistoryListItem *)item(currentItem()))->action();
   if (curr) {  // curr is an Action
      if (advanced == 1) { // curr is the "previous" Action
         if (curr->stepBackward() == curr->undoStep()) { // curr is done
            stopTimer();
            advanced = 0;
         }
      }
      else { // curr is not the "previous" Action
         if (currentItem() <= 0) { // curr is the first Action
            stopTimer();
            advanced = 0;
         } 
         else { // curr is not the first Action
            curr->preUndo();
            if (curr->isPermissible()) { // curr is ok to undo
               if (curr->stepBackward() == curr->firstStep()) { // curr is done 
                  setCurrentItem(currentItem() - 1);
                  game->undone.push(game->done.pop());
                  advanced = 1;
               }
            }
            else { // prev is impermissible or not an Action
               stopTimer();
               advanced = 0;   
            }
         }
      }
   }
   else { // curr is not an Action
      if (advanced == 1) { // curr is the "previous" action
         stopTimer();
         advanced = 0;
      }
      else { // curr is the current Action
         setCurrentItem(currentItem() - 1);
         advanced = 1;
      }
   }
}


void HistoryListBox::stepForwardOne() {
   static int advanced = 0;

   Action *curr = ((HistoryListItem *)item(currentItem()))->action();
   if (curr) {  // curr is an Action
      if (advanced == 1) { // curr is the "next" Action
         if (curr->stepForward() == curr->commitStep()) { // curr is done
            stopTimer();
            advanced = 0;
         }
      }
      else { // curr is not the "next" Action
         if (currentItem() >= (int)count() - 1) { // curr is the last Action
            stopTimer();
            advanced = 0;
         } 
         else { // curr is not the last Action
            Action *next =
             ((HistoryListItem *)item(currentItem() + 1))->action();
            if (!next || next->isPermissible()) { // next is permissible
               if (curr->stepForward() == curr->lastStep()) { // curr is done 
                  setCurrentItem(currentItem() + 1);
                  game->done.push(game->undone.pop());
                  advanced = 1;
               }
            }
            else { // next is impermissible or not an Action
               stopTimer();
               advanced = 0;   
            }
         }
      }
   }
   else { // curr is not an Action
      if (advanced == 1) { // curr is the "next" action
         stopTimer();
         advanced = 0;
      }
      else { // curr is the current Action
         setCurrentItem(currentItem() + 1);
         advanced = 1;
      }
   }
}


void HistoryListBox::logAction(Action *a) {
   // clear the undone stack
   while (!game->undone.isEmpty()) delete game->undone.pop();
   
   // remove items after current item
   while (currentItem() < (int)count() - 1) removeItem(count() - 1);

   // create the new item 
   QListBoxItem *i = new HistoryListItem(a->name(), a);
   insertItem(i);

   setCurrentItem(count()-1);
}


void HistoryListBox::unlogAction(Action *) { 
   removeItem(count()-1);
   setCurrentItem(count()-1);
}


void HistoryListBox::abortAction(Action *) {
   removeItem(count()-1);
   setCurrentItem(count()-1);
}


//
// HistoryDialog
//
HistoryDialog::HistoryDialog(GameObj *g, QWidget *parent, const char *name,
 WFlags f) : TabPage(QIconSet(), "History", parent, name, f) {
   QVBoxLayout *vl = (QVBoxLayout *)layout();

   // history list
   lb = new HistoryListBox(g, this);
   lb->setSelectionMode(QListBox::Single);
   vl->addWidget(lb);

   // VCR buttons
   QHBoxLayout *bl = new QHBoxLayout(vl);
   bl->setSpacing(0);
   bl->addStretch();

   QPushButton *firstButton = new QPushButton(this);
   firstButton->setPixmap(QPixmap((const char **)first_xpm));
   firstButton->setFixedSize(27, 27);
   QToolTip::add(firstButton, "First");
   connect(firstButton, SIGNAL(clicked()), lb, SLOT(first()));
   bl->addWidget(firstButton);

   QPushButton *rewindButton= new QPushButton(this);
   rewindButton->setPixmap(QPixmap((const char **)rewind_xpm));
   rewindButton->setFixedSize(27, 27);
   QToolTip::add(rewindButton, "Rewind");
   connect(rewindButton, SIGNAL(clicked()), lb, SLOT(rewind()));
   bl->addWidget(rewindButton);

   QPushButton *prevButton = new QPushButton(this);
   prevButton->setPixmap(QPixmap((const char **)prev_xpm));
   prevButton->setFixedSize(27, 27);
   QToolTip::add(prevButton, "Previous");
   connect(prevButton, SIGNAL(clicked()), lb, SLOT(prev()));
   bl->addWidget(prevButton);   

   QPushButton *stopButton = new QPushButton(this);
   stopButton->setPixmap(QPixmap((const char **)stop_xpm));
   stopButton->setFixedSize(27, 27);
   QToolTip::add(stopButton, "Stop");
   connect(stopButton, SIGNAL(clicked()), lb, SLOT(stop()));
   bl->addWidget(stopButton);   

   QPushButton *nextButton = new QPushButton(this);
   nextButton->setPixmap(QPixmap((const char **)next_xpm));
   nextButton->setFixedSize(27, 27);
   QToolTip::add(nextButton, "Next");
   connect(nextButton, SIGNAL(clicked()), lb, SLOT(next()));
   bl->addWidget(nextButton);   

   QPushButton *playButton = new QPushButton(this);
   playButton->setPixmap(QPixmap((const char **)play_xpm));
   playButton->setFixedSize(27, 27);
   QToolTip::add(playButton, "Play");
   connect(playButton, SIGNAL(clicked()), lb, SLOT(play()));
   bl->addWidget(playButton);   

   QPushButton *lastButton = new QPushButton(this);
   lastButton->setPixmap(QPixmap((const char **)last_xpm));
   lastButton->setFixedSize(27, 27);
   QToolTip::add(lastButton, "Last");
   connect(lastButton, SIGNAL(clicked()), lb, SLOT(last()));
   bl->addWidget(lastButton);

   bl->addStretch();
}


void HistoryDialog::finalize() { lb->finalize(); }


void HistoryDialog::prev() { lb->prev(); }


void HistoryDialog::next() { lb->next(); }


void HistoryDialog::logAction(Action *a) { lb->logAction(a); }


void HistoryDialog::unlogAction(Action *a) { lb->unlogAction(a); }


void HistoryDialog::abortAction(Action *a) { lb->abortAction(a); }
