// $Id: loadhandlers.cpp,v 1.50 2004/08/23 15:42:36 uckelman Exp $
/* Board -- a graphical frontend for boardgames
 * Copyright (C) 2002, 2003, 2004 Joel Uckelman
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <qcolor.h>
#include <qfiledialog.h>
#include <qinputdialog.h>

#include "tokendialogs.h"
#include "game.h"
#include "groupdialogs.h"
#include "loadhandlers.h"
#include "main.h"
#include "player.h"
#include "surfacedialogs.h"

extern Main *MW;


//
// BaseHandler
//
BaseHandler::BaseHandler(GameObj *game, const QString &file) :
 XmlHandler(file), g(game) {
}


BaseHandler::ErrorType BaseHandler::validateUniqueGroupable(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a unique Groupable id?
   if (g->gid2g.contains(tmp)) return AttributeNotUnique;

   val = tmp;
   return Ok;
}


BaseHandler::ErrorType BaseHandler::validateUniquePixmap(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a unique Groupable id?
   if (g->tPics.contains(tmp)) return AttributeNotUnique;

   val = tmp;
   return Ok;
}


BaseHandler::ErrorType BaseHandler::validateUniqueTokenType(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a unique TokenTypeid?
   if (g->tTypes.contains(tmp)) return AttributeNotUnique;

   val = tmp;
   return Ok;
}


BaseHandler::ErrorType BaseHandler::validateUniqueSurface(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a unique Surface id?
   if (g->sid2s.contains(tmp)) return AttributeNotUnique;

   val = tmp;
   return Ok;
}


BaseHandler::ErrorType BaseHandler::validateMatchGroup(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a token id?
   if (!g->gid2g.contains(tmp) || !dynamic_cast<Group *>(g->gid2g[tmp]))
    return AttributeNotMatched;

   val = tmp;
   return Ok;
}


BaseHandler::ErrorType BaseHandler::validateMatchPixmap(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a pixmap id?
   if (!g->tPics.contains(tmp)) return AttributeNotMatched;

   val = tmp;
   return Ok;
}


BaseHandler::ErrorType BaseHandler::validateMatchPlayer(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a token id?
   if (!g->gid2g.contains(tmp) || !dynamic_cast<Player *>(g->gid2g[tmp]))
    return AttributeNotMatched;

   val = tmp;
   return Ok;
}


BaseHandler::ErrorType BaseHandler::validateMatchSide(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val,
 unsigned int type) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a side in type?
   TokenPixmapList pl = g->tTypes[type]->side;
   for (TokenPixmapList::const_iterator i = pl.constBegin();
    i != pl.constEnd(); ++i) {
      if ((*i).id() == tmp) {
         val = tmp;
         return Ok;
      }
   }

   return AttributeNotMatched;
}


BaseHandler::ErrorType BaseHandler::validateMatchToken(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a token id?
   if (!g->gid2g.contains(tmp) || !dynamic_cast<Token *>(g->gid2g[tmp]))
    return AttributeNotMatched;

   val = tmp;
   return Ok;
}


BaseHandler::ErrorType BaseHandler::validateMatchTokenType(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a token id?
   if (!g->tTypes.contains(tmp)) return AttributeNotMatched;

   val = tmp;
   return Ok;
}


BaseHandler::ErrorType BaseHandler::validateMatchSurface(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a token id?
   if (!g->sid2s.contains(tmp) || !dynamic_cast<Surface *>(g->sid2s[tmp]))
    return AttributeNotMatched;

   val = tmp;
   return Ok;
}


//
// GameboxHandler
//
GameboxHandler::GameboxHandler(GameObj *game, const QString &file,
 bool loadScen) : BaseHandler(game, file), loadScenarios(loadScen) {
}


bool GameboxHandler::startElement(const QString &, const QString &,
 const QString &qName, const QXmlAttributes &atts) {
   // get attributes
   QMap<QString,QString> attrib = attributes(atts);
   bool ok = false;

   // process the element
   if (qName == "gamebox") ok = parseGamebox(qName, attrib);
   else if (qName == "pix") ok = parsePix(qName, attrib);
   else if (qName == "pixmaps") ok = parsePixmaps(qName, attrib);
   else if (qName == "scenario") ok = parseScenario(qName, attrib);
   else if (qName == "scenarios") ok = parseScenarios(qName, attrib);
   else if (qName == "sheet") ok = parseSheet(qName, attrib);
   else if (qName == "side") ok = parseSide(qName, attrib); 
   else if (qName == "title") ok = parseTitle(qName, attrib);
   else if (qName == "type") ok = parseType(qName, attrib);
   else if (qName == "types") ok = parseTypes(qName, attrib);
   else {
      // fail if element is unrecognized
      return setError(ElementUnrecognized, qName);
   }
     
   elementStack.push(qName);
   prevSibling = QString::null;
   return ok;
}


bool GameboxHandler::endElement(const QString &, const QString &,
 const QString &) {
   if (elementStack.top() == "scenarios" && loadScenarios) {
      // prompt user for scenario
      bool ok = false;
      QString scenario = QInputDialog::getItem("Choose a Scenario",
       "Scenario:", scenarios, 0, false, &ok, MW);
      if (!ok || scenario.isNull()) return false;

      g->scenariofile = s2f[scenario];

      if (!spawnScenarioHandler(scenario)) return false;
   }
   else if (elementStack.top() == "type") {
      g->tTypes[def->type] = def;
      
      // add token to All list
      g->alldialog->append(def);
   }

   prevSibling = elementStack.pop();
   return true;
}


bool GameboxHandler::characters(const QString &ch) {
   if (elementStack.top() == "title") g->title = ch;
   else if (elementStack.top() == "scenario") {
      scenarios.append(ch);
      s2f[ch] = scenariofile;
   }

   return true;
}


bool GameboxHandler::parseGamebox(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if gamebox is not the root element
   if (!elementStack.isEmpty() || !prevSibling.isNull())
    return setError(ElementNotRoot, "gamebox");

   // fail if version attribute is missing
   if (!attrib.contains("version"))
    return setError(AttributeMissing, qName, "version");
      
   QString boxver = attrib["version"];

   // fail if g->boxver is set and does not equal boxver
   if (g->boxver.isNull()) g->boxver = boxver;
   else if (g->boxver != boxver) 
    return setError(ValuesNotEqual, "gamebox version", boxver,
     "save or scenario version", g->boxver);
      
   // fail if id attribute is missing
   if (!attrib.contains("id")) return setError(AttributeMissing, qName, "id");

   QString boxid = attrib["id"];
   
   // fail if g->boxid is set and does not equal boxid
   if (g->boxid.isNull()) g->boxid = boxid;
   else if (g->boxid != boxid)
    return setError(ValuesNotEqual, "gamebox id", boxid,
     "save or scenario id", g->boxid);

   // fail if requires attribute is missing
   if (!attrib.contains("requires"))
    return setError(AttributeMissing, qName, "requires");
      
   QString requires = attrib["requires"];

   return true;
}


bool GameboxHandler::parsePix(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not a sheet element
   if (elementStack.top() != "sheet")
    return setError(ElementBadParent, qName, "sheet");
      
   // fail if previous sibling exists and is not a pix element
   if (!prevSibling.isNull() && prevSibling != "pix")
    return setError(ElementBadSibling, qName, qName);

   ErrorType ret;
   int x, y, w, h;
   unsigned int id;

   if ((ret = validateInt(attrib, "x", x)) != Ok)
    return setError(ret, qName, "x");
   if ((ret = validateInt(attrib, "y", y)) != Ok)
    return setError(ret, qName, "y");
   if ((ret = validateInt(attrib, "w", w)) != Ok)
    return setError(ret, qName, "w");
   if ((ret = validateInt(attrib, "h", h)) != Ok)
    return setError(ret, qName, "h");
   if ((ret = validateUniquePixmap(attrib, "id", id)) != Ok)
    return setError(ret, qName, "id");

   // cut token image from sheet as QImage
   QImage tmp = sheet.copy(x, y, w, h);
   
   // convert token QImage to TokenPixmap
   TokenPixmap pix(id);
   pix.convertFromImage(tmp);

   // put TokenPixmap in id map
   g->tPics[id] = pix;

   return true;
}


bool GameboxHandler::parsePixmaps(const QString &qName,
 QMap<QString,QString> &) {
   // fail if parent is not a gamebox element
   if (elementStack.top() != "gamebox")
    return setError(ElementBadParent, qName, "gamebox"); 
      
   // fail if previous sibling is not a title element
   if (prevSibling != "title") 
    return setError(ElementBadSibling,  "pixmaps", "title");

   return true;
}


bool GameboxHandler::parseScenario(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not a scenarios element
   if (elementStack.top() != "scenarios")
    return setError(ElementBadParent, qName, "scenarios");
      
   // fail if previous sibling exists and is not a scenario element
   if (!prevSibling.isNull() && prevSibling != "scenario")
    return setError(ElementBadSibling, qName, qName);

   // fail if name attribute is missing
   if (!attrib.contains("file"))
    return setError(AttributeMissing, qName, "file");

   scenariofile = attrib["file"];
   scenariofile.setFile(g->boxfile.dirPath(), scenariofile.fileName());
      
   // scenario name is set in GameboxHandler::characters()
   return true;
}


bool GameboxHandler::parseScenarios(const QString &qName,
 QMap<QString,QString> &) {
   // fail if parent is not a gamebox element
   if (elementStack.top() != "gamebox")
    return setError(ElementBadParent, qName, "gamebox");
      
   // fail if previous sibling is not a types element
   if (prevSibling != "types")
    return setError(ElementBadSibling, "scenarios", "types");

   return true;
}


bool GameboxHandler::parseSheet(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not a pixmaps element
   if (elementStack.top() != "pixmaps")
    return setError(ElementBadParent, qName, "pixmaps");
      
   // fail if previous sibling exists and is not a sheet element
   if (!prevSibling.isNull() && prevSibling != "sheet")
    return setError(ElementBadSibling, qName, qName);

   // fail if file attribute is missing
   if (!attrib.contains("file"))
    return setError(AttributeMissing, qName, "file");
      
   sheetfile.setFile(g->boxfile.dirPath(), attrib["file"]);

   // fail if sheet cannot be loaded
   if (!sheet.load(sheetfile.filePath()))
    return setError(FileReadFailed, sheetfile.absFilePath());

   return true;
}


bool GameboxHandler::parseSide(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not a type element
   if (elementStack.top() != "type")
    return setError(ElementBadParent, qName, "type");
      
   // fail if previous sibling exists and is not a side element
   if (!prevSibling.isNull() && prevSibling != "side")
    return setError(ElementBadSibling, qName, qName);

   ErrorType ret;
   unsigned int id;
   
   if ((ret = validateMatchPixmap(attrib, "id", id)) != Ok)
    return setError(ret, qName, "id");   
      
   def->side.append(g->tPics[id]);

   return true;
}


bool GameboxHandler::parseTitle(const QString &qName,
 QMap<QString,QString> &) {
   // fail if parent is not a gamebox element
   if (elementStack.top() != "gamebox" || !prevSibling.isNull())
    return setError(ElementBadParent, qName, "gamebox");

   // fail if not the first child
   if (!prevSibling.isNull())
    return setError(ElementBadSibling, qName, "(none)");

   // title text is set in GameboxHandler::characters()
   return true;
}


bool GameboxHandler::parseType(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not a types element
   if (elementStack.top() != "types")
    return setError(ElementBadParent, qName, "types");
      
   // fail if previous sibling exists and is not a type element
   if (!prevSibling.isNull() && prevSibling != "type")
    return setError(ElementBadSibling, qName, qName);
   
   ErrorType ret;
   unsigned int id;

   if ((ret = validateUniqueTokenType(attrib, "id", id)) != Ok)
    return setError(ret, qName, "id");

   // fail if name attribute is missing
   if (!attrib.contains("name"))
    return setError(AttributeMissing, qName, "name"); 

   QString name = attrib["name"];
      
   // create new type
   def = new TokenType;
   def->type = id;
   def->name = name;

   return true;
}


bool GameboxHandler::parseTypes(const QString &qName,
 QMap<QString,QString> &) {
   // fail if parent is not a gamebox element
   if (elementStack.top() != "gamebox")
    return setError(ElementBadParent, qName, "gamebox");
      
   // fail if previous sibling is not a pixmaps element
   if (prevSibling != "pixmaps")
    return setError(ElementBadSibling, "types", "pixmaps");

   return true;
}


bool GameboxHandler::spawnScenarioHandler(const QString &scenario) {
   QString data;
   
   // fail if scenario cannot be read
   if (!MW->readFile(g->scenariofile, data))
    return setError(FileReadFailed, g->scenariofile.absFilePath());

   // parse the scenario
   QXmlSimpleReader reader;
   ScenarioHandler handler(g, g->scenariofile.fileName(), scenario);
   reader.setContentHandler(&handler);
   reader.setErrorHandler(&handler);
   QXmlInputSource source;
   source.setData(data);
   return reader.parse(&source);
}


//
// SetupHandler
//
SetupHandler::SetupHandler(GameObj *game, const QString &file) :
 BaseHandler(game, file) {
}


bool SetupHandler::endElement(const QString &, const QString &,
 const QString &) { 
   if (elementStack.top() == "entities") {
      // set state
      QMap<QString,QString> attrib;

      for (QMapConstIterator<Group *, QMap<unsigned int, State> > i =
       state.begin(); i != state.end(); ++i) {
         Group *group = i.key();
         
         for (QMapConstIterator<unsigned int, State> j = i.data().begin();
          j != i.data().end(); ++j) {
            attrib["pid"] = QString::number(j.key());

            ErrorType ret;
            unsigned int pid;

            // FIXME: error message is off---permissions is not an element
            if ((ret = validateMatchPlayer(attrib, "pid", pid)) != Ok)
             return setError(ret, "permissions", "pid");
      
            Player *p = (Player *)g->gid2g[pid];

            // set permissions
            group->setPermissions(p, j.data().perms);
         }
      }

      // set drawing
      for (QMapConstIterator<Player *, QMap<unsigned int, Drawing> > i =
       drawing.begin(); i != drawing.end(); ++i) {
         Player *player = i.key();
         
         for (QMapConstIterator<unsigned int, Drawing> j = i.data().begin();
          j != i.data().end(); ++j) {
            attrib["pid"] = QString::number(j.key());

            ErrorType ret;
            unsigned int pid;

            if ((ret = validateMatchPlayer(attrib, "pid", pid)) != Ok)
             return setError(ret, "", "pid");
      
            Player *p = (Player *)g->gid2g[pid];

            player->setDrawingWidth(p, j.data().width);
            player->setDrawingMargin(p, j.data().margin);
            player->setDrawingColor(p, QColor(j.data().color));
         }
      }
   }
  
   if (elementStack.top() == "entities" ||
       elementStack.top() == "group" ||
       elementStack.top() == "player") groupStack.pop();

   prevSibling = elementStack.pop();
   return true;
}


bool SetupHandler::parseDrawing(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not a player element
   if (elementStack.top() != "player")
    return setError(ElementBadParent, qName, "player");

   ErrorType ret;
   unsigned int pid;
   Drawing d;

   if ((ret = validateUInt(attrib, "pid", pid)) != Ok)
    return setError(ret, qName, "pid");
   if ((ret = validateUInt(attrib, "width", d.width)) != Ok)
    return setError(ret, qName, "width");
   if ((ret = validateUInt(attrib, "margin", d.margin)) != Ok)
    return setError(ret, qName, "margin");
  
   d.color = attrib["color"];

   drawing[(Player *)(groupStack.top())][pid] = d;

   return true;
}


bool SetupHandler::parseEntities(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if previous sibling is not a surfaces element
   if (prevSibling != "surfaces")
    return setError(ElementBadSibling, qName, "surfaces");

   unsigned int id;
   ErrorType ret;
   if ((ret = validateUniqueGroupable(attrib, "id", id)) != Ok)
    return setError(ret, qName, "id");

   // create root group
   g->root = new RootGroup(g, id);
   g->groupsdialog->insert(g->root);
   g->gid2g[id] = g->root;
        
   // push root group onto the stack
   groupStack.push(g->root);

   // add root to group list
   groups << id;

   return true;
}


bool SetupHandler::parseGroup(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not a group, entities, or player element
   if (elementStack.top() != "entities" &&
       elementStack.top() != "group" &&
       elementStack.top() != "player") 
    return setError(ElementBadParent, "group", "entities, group, player");

   ErrorType ret;
   unsigned int id;
   if ((ret = validateUniqueGroupable(attrib, "id", id)) != Ok) 
    return setError(ret, qName, "id");

   // fail if name attribute is missing
   if (!attrib.contains("name"))
    return setError(AttributeMissing, qName, "name");

   QString name = attrib["name"];
   
   // create a new group
   Group *gr = new Group(g, name, groupStack.top(), id);

   // add to the group tree
   g->groupsdialog->insert(gr);

   // associate id with Groupable pointer
   g->gid2g[id] = gr;

   // push onto the stack
   groupStack.push(gr);

   // add id to group list
   groups << id;

   return true;
}


bool SetupHandler::parsePlayer(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an entities  element
   if (elementStack.top() != "entities")
    return setError(ElementBadParent, qName, "entities");
      
   // fail if previous sibling exists and is not a group, player,
   // or token element
   if (!prevSibling.isNull() &&
       prevSibling != "group" &&
       prevSibling != "player" &&
       prevSibling != "token")
    return setError(ElementBadSibling, qName, "group, player, token");

   ErrorType ret;
   unsigned int id;
   if ((ret = validateUniqueGroupable(attrib, "id", id)) != Ok)
    return setError(ret, qName, "id");

   // fail if name attribute is missing
   if (!attrib.contains("name"))
    return setError(AttributeMissing, qName, "name");

   QString name = attrib["name"];

   // fail if password attribute is missing
   if (!attrib.contains("password"))
    return setError(AttributeMissing, qName, "password");

   QString password = attrib["password"];

   // create a new player
   Player *pl = new Player(g, name, password, id);
 
   // add to the group tree
   g->groupsdialog->insert(pl);
 
   // associate id with Groupable pointer
   g->gid2g[id] = pl;

   // push onto the stack
   groupStack.push(pl);

   // add id to player, group lists
   players << id;
   groups << id;

   return true;
}


bool SetupHandler::parseToken(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not a entities, group, or player element
   if (elementStack.top() != "entities" &&
       elementStack.top() != "group" &&
       elementStack.top() != "player")
    return setError(ElementBadParent, qName, "entities, group, player"); 
      
   ErrorType ret;
   unsigned int id, surface, type, up;
   int x, y, z, theta;

   if ((ret = validateUniqueGroupable(attrib, "id", id)) != Ok)
    return setError(ret, qName, "id");
   if ((ret = validateMatchSurface(attrib, "surface", surface)) != Ok)
    return setError(ret, qName, "surface");
   if ((ret = validateMatchTokenType(attrib, "type", type)) != Ok)
    return setError(ret, qName, "type");
   if ((ret = validateInt(attrib, "x", x)) != Ok)
    return setError(ret, qName, "x");
   if ((ret = validateInt(attrib, "y", y)) != Ok)
    return setError(ret, qName, "y");
   if ((ret = validateInt(attrib, "z", z)) != Ok)
    return setError(ret, qName, "z");
   if ((ret = validateInt(attrib, "t", theta)) != Ok)
    return setError(ret, qName, "t");
   if ((ret = validateMatchSide(attrib, "up", up, type)) != Ok)
    return setError(ret, qName, "up");
      
   // fail if name attribute is missing
   if (!attrib.contains("name"))
    return setError(AttributeMissing, qName, "name");

   QString name = attrib["name"];
   
   // make token                     
   Token *t = new Token(g, groupStack.top(),
    g->sid2s[surface]->canvas(), x, y, z, theta, name, type, up, id);

   // add token to the group tree
   g->groupsdialog->insert(t);

   // add token to In Play list
   g->inplaydialog->insert(t);
 
   // associate token id with Groupable pointer   
   g->gid2g[id] = t;

   // add token id to the id list
   tokens << id;
            
   return true;
}


bool SetupHandler::parseState(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not a group or player element
   if (elementStack.top() != "group" &&
       elementStack.top() != "player")
    return setError(ElementBadParent, qName, "group, player");

   ErrorType ret;
   unsigned int pid;
   State s;

   if ((ret = validateUInt(attrib, "pid", pid)) != Ok)
    return setError(ret, qName, "pid");

   s.perms = attrib["permissions"];

   state[groupStack.top()][pid] = s;

   return true;
}


bool SetupHandler::parseSurface(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not a surfaces element
   if (elementStack.top() != "surfaces")
    return setError(ElementBadParent, qName, "surfaces");
      
   // fail if previous sibling exists and is not a surface element
   if (!prevSibling.isNull() && prevSibling != "surface")
    return setError(ElementBadSibling, qName, qName);

   ErrorType ret;
   unsigned int id;

   if ((ret = validateUniqueSurface(attrib, "id", id)) != Ok)
    return setError(ret, qName, "id");

   // fail if name attribute is missing
   if (!attrib.contains("name"))
    return setError(AttributeMissing, qName, "name"); 

   QString name = attrib["name"];
      
#ifdef Q_WS_WIN
   // Load background as QImage on Windows.
   QImage bg;
#else
   // Load background as QPixmap on all sane platforms.
   QPixmap bg;
#endif
   QFileInfo bf;
   if (attrib.contains("background")) {
      bf = attrib["background"];
      bf.setFile(g->boxfile.dirPath(), bf.fileName());

      if (!bg.load(bf.filePath()))
       return setError(FileReadFailed, bf.absFilePath());
   }
      
   // create surface
   QCanvas *c = new QCanvas;
   Surface *t = new Surface(g, c, name, bf, bg, id);
   t->setCaption(g->title + " - " + t->surfaceName());
   g->c2s[c] = t;
   g->sid2s[id] = t;            

   g->surfacesdialog->insert(t);

   surfaces << id;
 
   return true;
}


bool SetupHandler::parseSurfaces(const QString &, QMap<QString,QString> &) {
   return true;
}


bool SetupHandler::spawnGameboxHandler() {
   // find a gamebox if the specified one is problematic
   while (!MW->fileOk(g->boxfile)) {
      // prompt for a gamebox            
      g->boxfile = QFileInfo(QFileDialog::getOpenFileName(".",
       "Gameboxes (*.box)", MW, "gamebox", "Select a Gamebox..."));

      // give up if user cancels
      if (g->boxfile.filePath().isEmpty())
       return setError(ErrorVerbatim, "Load cancelled");
   }

   QString data;
   
   // fail if gamebox cannot be read
   if (!MW->readFile(g->boxfile, data))
    return setError(FileReadFailed, g->boxfile.absFilePath());

   // parse the gamebox
   QXmlSimpleReader reader;
   GameboxHandler handler(g, g->boxfile.fileName(), false);
   reader.setContentHandler(&handler);
   reader.setErrorHandler(&handler);
   QXmlInputSource source;
   source.setData(data);
   return reader.parse(&source);
}


//
// SaveHandler
//
SaveHandler::SaveHandler(GameObj *game, const QString &file) :
 SetupHandler(game, file) {
}


bool SaveHandler::startElement(const QString &, const QString &,
 const QString &qName, const QXmlAttributes &atts) {
   // get attributes
   QMap<QString,QString> attrib = attributes(atts);
   bool ok = false;

   if (qName == "actions") ok = parseActions(qName, attrib);
   else if (qName == "combo") ok = parseActionCombo(qName, attrib);
   else if (qName == "drawing") ok = parseDrawing(qName, attrib);
   else if (qName == "entities") ok = parseEntities(qName, attrib);
   else if (qName == "gcreate") ok = parseActionGroupCreate(qName, attrib);
   else if (qName == "ghide") ok = parseActionGroupHide(qName, attrib);
   else if (qName == "gregroup") ok = parseActionGroupRegroup(qName, attrib);
   else if (qName == "grename") ok = parseActionGroupRename(qName, attrib);
   else if (qName == "group") ok = parseGroup(qName, attrib);
   else if (qName == "gshow") ok = parseActionGroupShow(qName, attrib);
   else if (qName == "gtrash") ok = parseActionGroupTrash(qName, attrib);
   else if (qName == "pcreate") ok = parseActionPlayerCreate(qName, attrib);
   else if (qName == "prename") ok = parseActionPlayerRename(qName, attrib);
   else if (qName == "player") ok = parsePlayer(qName, attrib);
   else if (qName == "ptrash") ok = parseActionPlayerTrash(qName, attrib);
   else if (qName == "save") ok = parseSave(qName, attrib);
   else if (qName == "screate") ok = parseActionSurfaceCreate(qName, attrib);
   else if (qName == "srename") ok = parseActionSurfaceRename(qName, attrib);
   else if (qName == "state") ok = parseState(qName, attrib);
   else if (qName == "screate") ok = parseActionSurfaceCreate(qName, attrib);
   else if (qName == "strash") ok = parseActionSurfaceTrash(qName, attrib);
   else if (qName == "surface") ok = parseSurface(qName, attrib);
   else if (qName == "surfaces") ok = parseSurfaces(qName, attrib);
   else if (qName == "tclone") ok = parseActionTokenClone(qName, attrib);
   else if (qName == "tcreate") ok = parseActionTokenCreate(qName, attrib);
   else if (qName == "tflip") ok = parseActionTokenFlip(qName, attrib);
   else if (qName == "tinvert") ok = parseActionTokenInvert(qName, attrib);
   else if (qName == "tmove") ok = parseActionTokenMove(qName, attrib);
   else if (qName == "token") ok = parseToken(qName, attrib); 
   else if (qName == "tregroup") ok = parseActionTokenRegroup(qName, attrib);
   else if (qName == "trename") ok = parseActionTokenRename(qName, attrib);
   else if (qName == "trotate") ok = parseActionTokenRotate(qName, attrib);
   else if (qName == "tshuffle") ok = parseActionTokenShuffle(qName, attrib);
   else if (qName == "tstack") ok = parseActionTokenStack(qName, attrib);
   else if (qName == "ttrash") ok = parseActionTokenTrash(qName, attrib);
   else {
      // fail if element is unrecognized
      return setError(ElementUnrecognized, qName);
   }

   elementStack.push(qName);
   prevSibling = QString::null;
   return ok;
}


bool SaveHandler::endElement(const QString &namespaceURI,
 const QString &localName, const QString &qName) {
   // every combo MUST be listed here, otherwise undos will behave strangely
   if (!comboStack.isEmpty() &&
    (elementStack.top() == "combo" ||
     elementStack.top() == "gtrash" ||
     elementStack.top() == "ptrash" ||
     elementStack.top() == "strash" ||
     elementStack.top() == "tinvert" ||
     elementStack.top() == "tshuffle" ||
     elementStack.top() == "tstack"))
    comboStack.pop();

   return SetupHandler::endElement(namespaceURI, localName, qName);
}


bool SaveHandler::parseActions(const QString &qName,
 QMap<QString,QString> &) {
   // fail if parent is not a save element
   if (elementStack.top() != "save")
    return setError(ElementBadParent, qName, "save");

   // fail if previous sibling is not an entities element
   if (prevSibling != "entities")
    return setError(ElementBadSibling, qName, "entities");

   return true; 
}


bool SaveHandler::parseActionCombo(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo") 
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");

   ComboAction *a = new ComboAction(oid);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   // push the new combo onto the stack
   comboStack.push(a);

   return true;
}


bool SaveHandler::parseActionGroupCreate(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo")
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid, gid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateMatchPickledGroup(attrib, "gid", gid)) != Ok)
    return setError(ret, qName, "gid");
  
   // create the create action
   Action *a = new GroupCreateAction(oid, gid); 

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   // remove this id from the group list
   groups.remove(gid);

   return true;
}


bool SaveHandler::parseActionGroupHide(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo")
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid, gid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateMatchPickledGroup(attrib, "gid", gid)) != Ok)
    return setError(ret, qName, "gid");
 
   // create the hide action
   Action *a = new GroupHideAction(oid, gid);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   return true;
}


bool SaveHandler::parseActionGroupRegroup(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo")
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid, gid, pid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateMatchPickledGroup(attrib, "gid", gid)) != Ok)
    return setError(ret, qName, "gid");
   if ((ret = validateMatchPickledGroup(attrib, "pid", pid)) != Ok)
    return setError(ret, qName, "pid");
  
   // create the create action
   Action *a = new GroupRegroupAction(oid, gid, pid); 

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   return true;
}


bool SaveHandler::parseActionGroupRename(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo")
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid, gid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateMatchPickledGroup(attrib, "gid", gid)) != Ok)
    return setError(ret, qName, "gid");
 
   QString name = attrib["name"];
 
   // create the rename action
   Action *a = new GroupRenameAction(oid, gid, name);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   return true;
}


bool SaveHandler::parseActionGroupShow(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo")
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid, gid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateMatchPickledGroup(attrib, "gid", gid)) != Ok)
    return setError(ret, qName, "gid");
 
   // create the show action
   Action *a = new GroupShowAction(oid, gid);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   return true;
}


bool SaveHandler::parseActionGroupTrash(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions, combo, gtrash, or ptrash element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo" &&
       elementStack.top() != "gtrash" &&
       elementStack.top() != "ptrash") 
    return setError(ElementBadParent, qName, "actions, combo, gtrash, ptrash");

   ErrorType ret;
   unsigned int oid, gid, pid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateUniquePickledGroupable(attrib, "gid", gid)) != Ok)
    return setError(ret, qName, "gid");
   if ((ret = validateMatchPickledGroup(attrib, "pid", pid)) != Ok)
    return setError(ret, qName, "pid");

   QString name = attrib["name"];

   ComboAction *a = new GroupTrashAction(oid, gid, pid, name);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   // push the new combo onto the stack
   comboStack.push(a);

   // add id to the group id list
   groups << gid;

   return true;
}


bool SaveHandler::parseActionPlayerCreate(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo") 
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid, pid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateMatchPickledPlayer(attrib, "pid", pid)) != Ok)
    return setError(ret, qName, "pid");
  
   // create the create action
   Action *a = new PlayerCreateAction(oid, pid);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   // remove this id from the player, group lists
   players.remove(pid);
   groups.remove(pid);

   return true;
}


bool SaveHandler::parseActionPlayerRename(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo") 
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid, pid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateMatchPickledPlayer(attrib, "pid", pid)) != Ok)
    return setError(ret, qName, "pid");
 
   QString name = attrib["name"];
 
   // create the rename action
   Action *a = new PlayerRenameAction(oid, pid, name);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   return true;
}


bool SaveHandler::parseActionPlayerTrash(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions, combo, gtrash, or ptrash element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo")
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid, pid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateUniquePickledGroupable(attrib, "pid", pid)) != Ok)
    return setError(ret, qName, "pid");

   QString name = attrib["name"];
   QString pass = attrib["password"];

   ComboAction *a = new PlayerTrashAction(oid, pid, name, pass);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   // push the new combo onto the stack
   comboStack.push(a);

   // add id to the groups and players id lists
   groups << pid;
   players << pid;

   return true;
}


bool SaveHandler::parseActionSurfaceCreate(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo") 
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid, sid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateMatchPickledSurface(attrib, "sid", sid)) != Ok)
    return setError(ret, qName, "sid");
  
   // create the create action
   Action *a = new SurfaceCreateAction(oid, sid);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   // remove this id from the surface lists
   surfaces.remove(sid);

   return true;
}


bool SaveHandler::parseActionSurfaceRename(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo") 
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid, sid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateMatchPickledSurface(attrib, "sid", sid)) != Ok)
    return setError(ret, qName, "sid");
 
   QString name = attrib["name"];
 
   // create the rename action
   Action *a = new SurfaceRenameAction(oid, sid, name);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   return true;
}


bool SaveHandler::parseActionSurfaceTrash(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo")
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid, sid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateUniquePickledSurface(attrib, "sid", sid)) != Ok)
    return setError(ret, qName, "sid");

   QString name = attrib["name"];
   QString bg = attrib["background"];

   ComboAction *a = new SurfaceTrashAction(oid, sid, name, bg);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   // push the new combo onto the stack
   comboStack.push(a);

   // add id to the surfaces id list
   surfaces << sid;

   return true;
}


bool SaveHandler::parseActionTokenClone(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo")
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid, tid, cid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateMatchPickledToken(attrib, "tid", tid)) != Ok)
    return setError(ret, qName, "tid");
   if ((ret = validateMatchPickledToken(attrib, "cid", cid)) != Ok)
    return setError(ret, qName, "cid");
 
   // create the clone action
   TokenAction *a = new TokenCloneAction(oid, tid, cid);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   // remove this id from the token list
   tokens.remove(tid);

   return true;
}


bool SaveHandler::parseActionTokenCreate(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo")
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid, tid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateMatchPickledToken(attrib, "tid", tid)) != Ok)
    return setError(ret, qName, "tid");
  
   // create the create action
   TokenAction *a = new TokenCreateAction(oid, tid);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   // remove this id from the token list
   tokens.remove(tid);

   return true;
}


bool SaveHandler::parseActionTokenFlip(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo")
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid, tid;
   bool dir;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateMatchPickledToken(attrib, "tid", tid)) != Ok)
    return setError(ret, qName, "tid");
   if ((ret = validateBool(attrib, "dir", dir)) != Ok)
    return setError(ret, qName, "dir");
  
   // create the flip action
   TokenAction *a;
   if (dir) a = new TokenFlipRightAction(oid, tid);
   else a = new TokenFlipLeftAction(oid, tid);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   return true;
}


bool SaveHandler::parseActionTokenInvert(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo") 
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");

   ComboAction *a = new TokenInvertAction(oid);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   // push the new combo onto the stack
   comboStack.push(a);

   return true;
}


bool SaveHandler::parseActionTokenMove(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo" &&
       elementStack.top() != "tinvert" &&
       elementStack.top() != "tshuffle" &&
       elementStack.top() != "tstack") 
    return setError(ElementBadParent, qName,
     "actions, combo, tinvert, tshuffle, tstack");

   ErrorType ret;
   unsigned int oid, tid, sid;
   int ox, oy, oz;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateMatchPickledToken(attrib, "tid", tid)) != Ok)
    return setError(ret, qName, "tid");
   if ((ret = validateMatchPickledSurface(attrib, "sid", sid)) != Ok)
    return setError(ret, qName, "sid");
   if ((ret = validateInt(attrib, "ox", ox)) != Ok)
    return setError(ret, qName, "ox");
   if ((ret = validateInt(attrib, "oy", oy)) != Ok)
    return setError(ret, qName, "oy");
   if ((ret = validateInt(attrib, "oz", oz)) != Ok)
    return setError(ret, qName, "oz");

   // create the move action
   TokenAction *a = new TokenMoveAction(oid, tid, sid, ox, oy, oz);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   return true;
}


bool SaveHandler::parseActionTokenRegroup(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo")
    return setError(ElementBadParent, qName, "actions, combo");
   ErrorType ret;
   unsigned int oid, tid, gid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateMatchPickledToken(attrib, "tid", tid)) != Ok)
    return setError(ret, qName, "tid");
   if ((ret = validateMatchPickledGroup(attrib, "gid", gid)) != Ok)
    return setError(ret, qName, "gid");
 
   // create the regroup action
   TokenAction *a = new TokenRegroupAction(oid, tid, gid);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   return true;
}


bool SaveHandler::parseActionTokenRename(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo")
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid, tid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateMatchPickledToken(attrib, "tid", tid)) != Ok)
    return setError(ret, qName, "tid");
 
   QString name = attrib["name"];
 
   // create the rename action
   TokenAction *a = new TokenRenameAction(oid, tid, name);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   return true;
}


bool SaveHandler::parseActionTokenRotate(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo")
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid, tid;
   int dt;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateMatchPickledToken(attrib, "tid", tid)) != Ok)
    return setError(ret, qName, "tid");
   if ((ret = validateInt(attrib, "dt", dt)) != Ok)
    return setError(ret, qName, "dt");
  
   // create the rotate action
   // NOTE: -dt usually is counterclockwise; here -dt is clockwise,
   //  since these actions are setup to be UNdone.
   TokenAction *a;
   if (dt < 0) a = new TokenRotateCWAction(oid, tid, dt);
   else a = new TokenRotateCCWAction(oid, tid, dt);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   return true;
}


bool SaveHandler::parseActionTokenShuffle(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo") 
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");

   ComboAction *a = new TokenShuffleAction(oid);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   // push the new combo onto the stack
   comboStack.push(a);

   return true;
}


bool SaveHandler::parseActionTokenStack(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions or combo element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo") 
    return setError(ElementBadParent, qName, "actions, combo");

   ErrorType ret;
   unsigned int oid;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");

   ComboAction *a = new TokenStackAction(oid);

   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   // push the new combo onto the stack
   comboStack.push(a);

   return true;
}


bool SaveHandler::parseActionTokenTrash(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not an actions, combo, gtrash, ptrash,
   // or strash element
   if (elementStack.top() != "actions" &&
       elementStack.top() != "combo" &&
       elementStack.top() != "gtrash" &&
       elementStack.top() != "ptrash" &&
       elementStack.top() != "strash")
    return setError(ElementBadParent, qName,
     "actions, combo, gtrash, ptrash, strash");

   ErrorType ret;
   unsigned int oid, tid, sid, gid, type, up;
   int x, y, z, t;

   if ((ret = validateMatchPickledPlayer(attrib, "oid", oid)) != Ok)
    return setError(ret, qName, "oid");
   if ((ret = validateUniquePickledGroupable(attrib, "tid", tid)) != Ok)
    return setError(ret, qName, "tid");
   if ((ret = validateMatchPickledSurface(attrib, "sid", sid)) != Ok)
    return setError(ret, qName, "sid");
   if ((ret = validateMatchPickledGroup(attrib, "gid", gid)) != Ok)
    return setError(ret, qName, "gid");
   if ((ret = validateMatchTokenType(attrib, "type", type)) != Ok)
    return setError(ret, qName, "type");
   if ((ret = validateInt(attrib, "x", x)) != Ok)
    return setError(ret, qName, "x");
   if ((ret = validateInt(attrib, "y", y)) != Ok)
    return setError(ret, qName, "y");
   if ((ret = validateInt(attrib, "z", z)) != Ok)
    return setError(ret, qName, "z");
   if ((ret = validateInt(attrib, "t", t)) != Ok)
    return setError(ret, qName, "t");
   if ((ret = validateMatchSide(attrib, "up", up, type)) != Ok)
    return setError(ret, qName, "up");
      
   // fail if name attribute is missing
   if (!attrib.contains("name"))
    return setError(AttributeMissing, qName, "name");

   QString name = attrib["name"];
      
   // make trash action
   TokenAction *a = new TokenTrashAction(oid, tid, sid, gid, x, y, z, t, name,
    type, up);
 
   // prepend action to the stack or its combo
   if (comboStack.isEmpty()) g->done.prepend(a);
   else comboStack.top()->prepend(a);

   // add token id to the id list
   tokens << tid;
            
   return true;
}


bool SaveHandler::parseEntities(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not a save element
   if (elementStack.top() != "save")
    return setError(ElementBadParent, qName, "save");

   return SetupHandler::parseEntities(qName, attrib);
}


bool SaveHandler::parseSave(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if save is not the root element
   if (!elementStack.isEmpty() || !prevSibling.isNull())
    return setError(ElementNotRoot, "save");

   // fail if version attribute is missing
   if (!attrib.contains("version"))
    return setError(AttributeMissing, qName, "version");
     
   g->boxver = attrib["version"];
      
   // fail if id attribute is missing
   if (!attrib.contains("id")) return setError(AttributeMissing, qName, "id");

   g->boxid = attrib["id"];
      
   // fail if using attribute is missing
   if (!attrib.contains("using"))
    return setError(AttributeMissing, qName, "using");
      
   QString boardver = attrib["using"];

   // fail if gamebox attribute is missing
   if (!attrib.contains("gamebox"))
    return setError(AttributeMissing, qName, "gamebox");
      
   g->boxfile = QFileInfo(attrib["gamebox"]);

   return spawnGameboxHandler();
}


bool SaveHandler::parseSurfaces(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not a save element
   if (elementStack.top() != "save")
    return setError(ElementBadParent, qName, "save");

   // fail if not the first child
   if (!prevSibling.isNull())
    return setError(ElementBadSibling, qName, "(none)");

   return SetupHandler::parseSurfaces(qName, attrib);
}

   
BaseHandler::ErrorType SaveHandler::validateUniquePickledGroupable(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a unique Groupable id?
   if (tokens.find(tmp) != tokens.end() ||
       groups.find(tmp) != groups.end() ||
       players.find(tmp) != players.end())
    return AttributeNotUnique;

   val = tmp;
   return Ok;
}


BaseHandler::ErrorType SaveHandler::validateUniquePickledSurface(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a unique Surface id?
   if (surfaces.find(tmp) != surfaces.end()) return AttributeNotUnique;

   val = tmp;
   return Ok;
}


BaseHandler::ErrorType SaveHandler::validateMatchPickledGroup(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a token id?
   if (!groups.contains(tmp)) return AttributeNotMatched;

   val = tmp;
   return Ok;
}


BaseHandler::ErrorType SaveHandler::validateMatchPickledPlayer(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a token id?
   if (players.find(tmp) == players.end()) return AttributeNotMatched;

   val = tmp;
   return Ok;
}


BaseHandler::ErrorType SaveHandler::validateMatchPickledToken(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a token id?
   if (tokens.find(tmp) == tokens.end()) return AttributeNotMatched;

   val = tmp;
   return Ok;
}


BaseHandler::ErrorType SaveHandler::validateMatchPickledSurface(
 const QMap<QString,QString> &attr, const QString &key, unsigned int &val) {
   ErrorType ret;
   unsigned int tmp;

   // Get a good unsigned int
   if ((ret = validateUInt(attr, key, tmp)) != Ok) return ret;

   // Is tmp a token id?
   if (surfaces.find(tmp) == surfaces.end()) return AttributeNotMatched;

   val = tmp;
   return Ok;
}


//
// ScenarioHandler
//
ScenarioHandler::ScenarioHandler(GameObj *game, const QString &file,
 const QString &scenario) :  SetupHandler(game, file), scenarioName(scenario) {
}


bool ScenarioHandler::startElement(const QString &, const QString &,
 const QString &qName, const QXmlAttributes &atts) {
   // get attributes
   QMap<QString,QString> attrib = attributes(atts);
   bool ok = false;

   if (qName == "drawing") ok = parseDrawing(qName, attrib);
   else if (qName == "entities") ok = parseEntities(qName, attrib);
   else if (qName == "group") ok = parseGroup(qName, attrib);
   else if (qName == "player") ok = parsePlayer(qName, attrib);
   else if (qName == "scenario") ok = parseScenario(qName, attrib);
   else if (qName == "token") ok = parseToken(qName, attrib); 
   else if (qName == "state") ok = parseState(qName, attrib);
   else if (qName == "surface") ok = parseSurface(qName, attrib);
   else if (qName == "surfaces") ok = parseSurfaces(qName, attrib);
   else {
      // fail if element is unrecognized
      return setError(ElementUnrecognized, qName);
   }

   elementStack.push(qName);
   prevSibling = QString::null;
   return ok;
}


bool ScenarioHandler::parseEntities(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not a scenario element
   if (elementStack.top() != "scenario")
    return setError(ElementBadParent, qName, "scenario");
 
   return SetupHandler::parseEntities(qName, attrib);
}


bool ScenarioHandler::parseScenario(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if scenario is not the root element
   if (!elementStack.isEmpty() || !prevSibling.isNull())
    return setError(ElementNotRoot, "scenario");

   // fail if version attribute is missing
   if (!attrib.contains("version"))
    return setError(AttributeMissing, qName, "version");
     
   // fail if id attribute is missing
   if (!attrib.contains("id"))
    return setError(AttributeMissing, qName, "id");

   // fail if requires attribute is missing
   if (!attrib.contains("requires"))
    return setError(AttributeMissing, qName, "requires");
        
   QString requires = attrib["requires"];
   
   // fail if gamebox attribute is missing
   if (!attrib.contains("gamebox"))
    return setError(AttributeMissing, qName, "gamebox");
   
   // fail if name attribute is missing
   if (!attrib.contains("name"))
    return setError(AttributeMissing, qName, "name");
 
   QString name = attrib["name"];

   if (scenarioName.isNull()) {
      scenarioName = name;

      g->boxver = attrib["version"];
      g->boxid = attrib["id"];

      g->boxfile = QFileInfo(attrib["gamebox"]);
      g->boxfile.setFile(g->scenariofile.dirPath(), g->boxfile.fileName());

      if (!spawnGameboxHandler()) return false;    
   }
   else {
      QString boxver = attrib["version"];
   
      // fail if g->boxver does not equal boxver
      if (g->boxver != boxver)
       return setError(ValuesNotEqual, "scenario version", boxver,
        "gamebox version", g->boxver);
      
      QString boxid = attrib["id"];
       
      // fail if g->boxid does not equal boxid
      if (g->boxid != boxid)
       return setError(ValuesNotEqual, "scenario id", boxid,
        "gamebox id", g->boxid);
   
      QString boxfile = attrib["gamebox"];
      
      // fail if g->boxfile does not equal boxfile
      if (g->boxfile.fileName() != boxfile)
       return setError(ValuesNotEqual, "scenario boxfile", boxfile,
        "gamebox boxfile", g->boxfile.fileName());

      // fail if name does not equal scenario name
      if (name != scenarioName)
       return setError(ValuesNotEqual, "scenario name in scenario", name,
        "scenario name in gamebox", scenarioName);
   }

   return true;
}


bool ScenarioHandler::parseSurfaces(const QString &qName,
 QMap<QString,QString> &attrib) {
   // fail if parent is not a scenario element
   if (elementStack.top() != "scenario")
    return setError(ElementBadParent, qName, "scenario");

   // fail if not the first child
   if (!prevSibling.isNull())
    return setError(ElementBadSibling, qName, "(none)");

   return SetupHandler::parseSurfaces(qName, attrib);
}
