// $Id: loadhandlers.h,v 1.30 2004/08/23 15:42:36 uckelman Exp $
/* Board -- a graphical frontend for boardgames
 * Copyright (C) 2002, 2003, 2004 Joel Uckelman
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __LOADHANDLERS_H_
#define __LOADHANDLERS_H_

#include <qfileinfo.h>
#include <qimage.h>

#include "xmlhandler.h"

class GameObj;
class Group;
class SelectionAction;
struct TokenType;


class BaseHandler : public XmlHandler {
   public:
      BaseHandler(GameObj *game, const QString &file);

   protected:
      ErrorType validateUniqueGroupable(const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val);
      ErrorType validateUniquePixmap(const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val);
      ErrorType validateUniqueTokenType(const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val);
      ErrorType validateUniqueSurface(const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val);

      ErrorType validateMatchGroup(const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val);
      ErrorType validateMatchPixmap(const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val);
      ErrorType validateMatchPlayer(const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val);
      ErrorType validateMatchSide(const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val, unsigned int type);
      ErrorType validateMatchSurface(const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val);
      ErrorType validateMatchToken(const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val);
      ErrorType validateMatchTokenType(const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val);

      GameObj *g;
};


class GameboxHandler : public BaseHandler {
   public:
      GameboxHandler(GameObj *game, const QString &file, bool loadScen = true);
      virtual bool startElement(const QString &namespaceURI,
       const QString &localName, const QString &qName,
       const QXmlAttributes &atts);
      virtual bool endElement(const QString &namespaceURI,
       const QString &localName, const QString &qName);
      virtual bool characters(const QString &ch);

   private:
      bool parseGamebox(const QString &qName, QMap<QString,QString> &attrib);
      bool parsePix(const QString &qName, QMap<QString,QString> &attrib);
      bool parsePixmaps(const QString &qName, QMap<QString,QString> &attrib);
      bool parseScenario(const QString &qName, QMap<QString,QString> &attrib);
      bool parseScenarios(const QString &qName, QMap<QString,QString> &attrib);
      bool parseSheet(const QString &qName, QMap<QString,QString> &attrib);
      bool parseSide(const QString &qName, QMap<QString,QString> &attrib);
      bool parseTitle(const QString &qName, QMap<QString,QString> &attrib);
      bool parseType(const QString &qName, QMap<QString,QString> &attrib);
      bool parseTypes(const QString &qName, QMap<QString,QString> &attrib);
    
      bool spawnScenarioHandler(const QString &scenario);
  
      QFileInfo sheetfile;
      QImage sheet;

      TokenType *def;

      QStringList scenarios;
      QFileInfo scenariofile;
      QMap<QString,QFileInfo> s2f;

      bool loadScenarios;
};


class SetupHandler : public BaseHandler {
   public:
      SetupHandler(GameObj *game, const QString &file);
      virtual bool endElement(const QString &namespaceURI,
       const QString &localName, const QString &qName);
      
   protected:
      bool parseDrawing(const QString &qName, QMap<QString,QString> &attrib);
      bool parseEntities(const QString &qName, QMap<QString,QString> &attrib);
      bool parseGroup(const QString &qName, QMap<QString,QString> &attrib);
      bool parsePerms(const QString &qName, QMap<QString,QString> &attrib);
      bool parsePlayer(const QString &qName, QMap<QString,QString> &attrib);
      bool parseToken(const QString &qName, QMap<QString,QString> &attrib);
      bool parseState(const QString &qName, QMap<QString,QString> &attrib);
      bool parseSurface(const QString &qName, QMap<QString,QString> &attrib);
      bool parseSurfaces(const QString &qName, QMap<QString,QString> &attrib);

      bool spawnGameboxHandler();

      struct State {
         QString perms;
         bool visible;
      };

      struct Drawing {
         unsigned int width, margin;
         QString color; 
      };

      QValueStack<Group *> groupStack;
      QMap<Group *, QMap<unsigned int, State> > state;
      QMap<Player *, QMap<unsigned int, Drawing> > drawing;
      QValueList<unsigned int> groups, players, surfaces, tokens;
};


class SaveHandler : public SetupHandler {
   public:
      SaveHandler(GameObj *game, const QString &file);
      virtual bool startElement(const QString &namespaceURI,
       const QString &localName, const QString &qName,
       const QXmlAttributes &atts);
      virtual bool endElement(const QString &namespaceURI,
       const QString &localName, const QString &qName);

   private:
      bool parseActions(const QString &qName, QMap<QString,QString> &attrib);
      bool parseActionCombo(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionGroupCreate(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionGroupHide(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionGroupRegroup(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionGroupRename(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionGroupShow(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionGroupTrash(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionPlayerCreate(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionPlayerRename(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionPlayerTrash(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionSurfaceCreate(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionSurfaceRename(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionSurfaceTrash(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionTokenClone(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionTokenCreate(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionTokenFlip(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionTokenInvert(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionTokenMove(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionTokenRegroup(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionTokenRename(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionTokenRotate(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionTokenShuffle(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionTokenStack(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseActionTokenTrash(const QString &qName,
       QMap<QString,QString> &attrib);
      bool parseEntities(const QString &qName, QMap<QString,QString> &attrib);
      bool parseSave(const QString &qName, QMap<QString,QString> &attrib);
      bool parseSurfaces(const QString &qName, QMap<QString,QString> &attrib);

      ErrorType validateUniquePickledGroupable(
       const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val);
      ErrorType validateUniquePickledSurface(
       const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val);
      ErrorType validateMatchPickledGroup(const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val);
      ErrorType validateMatchPickledPlayer(const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val);
      ErrorType validateMatchPickledSurface(const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val);
      ErrorType validateMatchPickledToken(const QMap<QString,QString> &attr,
       const QString &key, unsigned int &val);

      QValueStack<ComboAction *> comboStack;
};


class ScenarioHandler : public SetupHandler {
   public:
      ScenarioHandler(GameObj *game, const QString &file,
       const QString &scenario = QString::null);
      virtual bool startElement(const QString &namespaceURI,
       const QString &localName, const QString &qName,
       const QXmlAttributes &atts);

   private:
      bool parseEntities(const QString &qName, QMap<QString,QString> &attrib);
      bool parseSurfaces(const QString &qName, QMap<QString,QString> &attrib);
      bool parseScenario(const QString &qName, QMap<QString,QString> &attrib);

      QString scenarioName;
};

#endif /* __LOADHANDLERS_H_ */
