// $Id: polyline.cpp,v 1.1 2004/09/21 15:03:14 uckelman Exp $
/* Board -- a graphical frontend for boardgames
 * Copyright (C) 2002, 2003, 2004 Joel Uckelman
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "game.h"
#include "polyline.h"
#include "util.h"


//
// PolylineSegment
//
PolylineSegment::PolylineSegment(QPoint p1, QPoint p2, QCanvas *c, Polyline *l)
 : QCanvasLine(c), pline(l) {
   setPoints(p1.x(), p1.y(), p2.x(), p2.y());
}


int PolylineSegment::rtti() const { return PolylineSegment::RTTI; }


//
// Polyline
//
Polyline::Polyline(GameObj *g, const QString &n, Group *gr, QCanvas *c,
 unsigned int i) : Groupable(g, n, gr, i), cv(c), dot(0) {
}


Polyline::~Polyline() {
   /*
   for (QValueList<PolylineSegment *>::const_iterator i = segments.begin();
    i != segments.end(); ++i) delete *i;
   */
}


void Polyline::pushPoint(QPoint p2) {
   if (segments.count()) {
      QPoint p1(segments.top()->endPoint());
      PolylineSegment *seg = new PolylineSegment(p1, p2, cv, this);
      seg->setPen(pen);
      segments.push(seg);
   }
   else if (dot) {
      QPoint p1(int(dot->x()), int(dot->y()));
      PolylineSegment *seg = new PolylineSegment(p1, p2, cv, this);
      seg->setPen(pen);
      segments.push(seg);
      delete dot;
      dot = 0;      
   }
   else {
      dot = new QCanvasEllipse(pen.width(), pen.width(), cv);
      dot->setBrush(QBrush(pen.color()));
      dot->setX(p2.x());
      dot->setY(p2.y());
   }
}


QPoint Polyline::popPoint() {
   QPoint p2;

   if (segments.count()) {
      PolylineSegment *seg = segments.pop();
      p2 = seg->endPoint();
      
      if (segments.count() == 1) {
         QPoint p1 = seg->startPoint();
         dot = new QCanvasEllipse(pen.width(), pen.width(), cv);
         dot->setBrush(QBrush(pen.color()));
         dot->setX(p1.x());
         dot->setY(p1.y());
      }

      delete seg;
   }
   else if (dot) {
      p2 = QPoint(int(dot->x()), int(dot->y()));
      delete dot;
      dot = 0;
   }

   return p2;
}


void Polyline::setPen(QPen p) { pen = p; }


void Polyline::show() {
   if (segments.count())
    for (QValueStack<PolylineSegment *>::const_iterator i = segments.begin();
     i != segments.end(); ++i) (*i)->show();
   else if (dot) dot->show();
}


void Polyline::hide() {
   if (segments.count())
    for (QValueStack<PolylineSegment *>::const_iterator i = segments.begin();
     i != segments.end(); ++i) (*i)->hide();
   else if (dot) dot->hide(); 
}


bool Polyline::isVisible() const {
   if (dot) return dot->isVisible();
   else if (segments.count()) return segments.top()->isVisible();
   else return false;
}


QCanvas *Polyline::canvas() const {
   if (segments.count() || dot) return cv;
   else return 0;
}


void Polyline::setCanvas(QCanvas *c) {
   if (segments.count())
    for (QValueStack<PolylineSegment *>::const_iterator i = segments.begin();
     i != segments.end(); ++i) (*i)->setCanvas(c);
   else if (dot) dot->setCanvas(c); 

   cv = c;
}


void Polyline::setX(double x) {
   if (segments.count()) {
      double ox = segments.top()->endPoint().x() - x;
      for (QValueStack<PolylineSegment *>::const_iterator i = segments.begin();
       i != segments.end(); ++i) (*i)->setX((*i)->x()+ox);
   }
   else if (dot) dot->setX(x);
}


void Polyline::setY(double y) {
   if (segments.count()) {
      double oy = segments.top()->endPoint().y() - y;
      for (QValueStack<PolylineSegment *>::const_iterator i = segments.begin();
       i != segments.end(); ++i) (*i)->setY((*i)->y()+oy);
   }
   else if (dot) dot->setY(y);
}


void Polyline::setZ(double z) {
   if (segments.count()) 
    for (QValueStack<PolylineSegment *>::const_iterator i = segments.begin();
     i != segments.end(); ++i) (*i)->setZ((*i)->z()+z);
   else if (dot) dot->setZ(z);
}


bool Polyline::permission(const Player *pl, unsigned int pm) const {
   return Parent ? Parent->permission(pl, pm) : true;
}


QDomElement Polyline::dumpXML(QDomDocument &doc) const {
   if (!dot && segments.isEmpty()) return QDomElement();

   QDomElement e = doc.createElement("pline");
   e.setAttribute("id", id());
   e.setAttribute("name", Name);
   e.setAttribute("surface", game->c2s[cv]->id());

   if (dot) {

      QDomElement d = doc.createElement("point");
      QPoint c = dot->boundingRect().center();

      d.setAttribute("x", c.x());
      d.setAttribute("y", c.y());
      d.setAttribute("z", dot->z());
      d.setAttribute("w", dot->width());
      d.setAttribute("color", rgb2str(dot->pen().color().rgb()));

      e.appendChild(d);
   }
   else {
      for (QValueStack<PolylineSegment *>::const_iterator i = segments.begin();
       i != segments.end(); ++i) {
         QDomElement d = doc.createElement("point");
         QPoint s = (*i)->startPoint();
      
         d.setAttribute("x", s.x());
         d.setAttribute("y", s.y());
         d.setAttribute("z", (*i)->z());
         d.setAttribute("w", (*i)->pen().width());
         d.setAttribute("color", rgb2str((*i)->pen().color().rgb()));

         e.appendChild(d);
      }

      // last point
      PolylineSegment *seg = segments.last();

      QDomElement d = doc.createElement("point");
      QPoint s = seg->endPoint();
      
      d.setAttribute("x", s.x());
      d.setAttribute("y", s.y());
      d.setAttribute("z", seg->z());
      d.setAttribute("w", seg->pen().width());
      d.setAttribute("color", rgb2str(seg->pen().color().rgb()));

      e.appendChild(d);
   }

   return e;
}
