// $Id: selectionhandler.cpp,v 1.1 2004/08/24 14:51:36 uckelman Exp $
/* Board -- a graphical frontend for boardgames
 * Copyright (C) 2002, 2003, 2004 Joel Uckelman
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "game.h"
#include "selectionhandler.h"
#include "tokendialogs.h"


//
// SelectionHandler
//
SelectionHandler::SelectionHandler(GameObj *g, QObject *parent,
 const char *name) : QObject(parent, name), game(g), antsTimer(this) {

   connect(this, SIGNAL(hasSelection(bool)), this, SLOT(startAnts(bool)));
   connect(&antsTimer, SIGNAL(timeout()), this, SLOT(moveAnts()));
}


unsigned int SelectionHandler::count() const { return selected.count(); }


QCanvasItemList SelectionHandler::selection() const { return selected; }


void SelectionHandler::assignSelection(QCanvasItemList l) {
   QValueList<QCanvas *> update;

   // deselect old selection
   for (QCanvasItemList::const_iterator i = selected.begin();
    i != selected.end(); ++i) {
      (*i)->setSelected(false);
      --refcount[(*i)->canvas()];

      if (update.find((*i)->canvas()) == update.end())
       update.append((*i)->canvas());      
   }

   game->selecteddialog->clear();

   // select new selection
   for (QCanvasItemList::const_iterator i = l.begin(); i != l.end(); ++i) {
      (*i)->setSelected(true);
      ++refcount[(*i)->canvas()];

      if (update.find((*i)->canvas()) == update.end())
       update.append((*i)->canvas());

      if ((*i)->rtti() == Token::RTTI)
       game->selecteddialog->insert((Token *)*i);
   }
  
   // notify listeners
   selected = l;
   emit hasSelection(selected.count());

   // update canvases
   for (QValueList<QCanvas *>::const_iterator i = update.begin();
    i != update.end(); ++i) (*i)->update();
}


void SelectionHandler::clearSelection() {
   assignSelection(QCanvasItemList());
}


void SelectionHandler::unionSelection(QCanvasItemList l) {
   QValueList<QCanvas *> update;

   // select the addend
   for (QCanvasItemList::const_iterator i = l.begin(); i != l.end(); ++i) {
      (*i)->setSelected(true);
      ++refcount[(*i)->canvas()];

      if (update.find((*i)->canvas()) == update.end())
       update.append((*i)->canvas());

      if ((*i)->rtti() == Token::RTTI)
       game->selecteddialog->insert((Token *)*i);
   }

   // notify listeners
   selected += l;
   emit hasSelection(selected.count());

   // update canvases
   for (QValueList<QCanvas *>::const_iterator i = update.begin();
    i != update.end(); ++i) (*i)->update();
}


void SelectionHandler::subtractSelection(QCanvasItemList l) {
   QValueList<QCanvas *> update;

   // deselect the minuend
   for (QCanvasItemList::const_iterator i = l.begin(); i != l.end(); ++i) {
      if (selected.find(*i) != selected.end()) {
         (*i)->setSelected(false);
         --refcount[(*i)->canvas()];

         if (update.find((*i)->canvas()) == update.end())
          update.append((*i)->canvas());

         if ((*i)->rtti() == Token::RTTI)
          game->selecteddialog->remove((Token *)*i);

         selected.remove(*i);
      }
   }

   // notify listeners
   emit hasSelection(selected.count());

   // update canvases
   for (QValueList<QCanvas *>::const_iterator i = update.begin();
    i != update.end(); ++i) (*i)->update();
}


void SelectionHandler::intersectSelection(QCanvasItemList l) {
   QValueList<QCanvas *> update;

   // deselect the nonintersection 
   for (QCanvasItemList::const_iterator i = l.begin(); i != l.end(); ++i) {
      if (selected.find(*i) == selected.end()) {
         (*i)->setSelected(false);
         --refcount[(*i)->canvas()];

         if (update.find((*i)->canvas()) == update.end())
          update.append((*i)->canvas());

         if ((*i)->rtti() == Token::RTTI)
          game->selecteddialog->remove((Token *)*i);

         selected.remove(*i);
      }
   }

   // notify listeners
   emit hasSelection(selected.count());

   // update canvases
   for (QValueList<QCanvas *>::const_iterator i = update.begin();
    i != update.end(); ++i) (*i)->update();
}


void SelectionHandler::startAnts(bool start) {
   if (start) antsTimer.start(100);
   else antsTimer.stop();
}


void SelectionHandler::moveAnts() {
   static unsigned int phase = 0;

   // advance the ants
   for (QCanvasItemList::const_iterator i = selected.begin();
    i != selected.end(); ++i) {
      if ((*i)->rtti() == Token::RTTI) ((Token *)*i)->advanceAnts(phase);
   }

   // update canvases
   for (QMap<QCanvas *, unsigned int>::const_iterator i = refcount.begin();
    i != refcount.end(); ++i) if (i.data()) i.key()->update();

   // it is 6 pixels from beginning of one ant to the beginning of the next
   phase = (phase + 1) % 6;
}
