// $Id: surface.cpp,v 1.2 2004/08/24 14:41:16 uckelman Exp $
/* Board -- a graphical frontend for boardgames
 * Copyright (C) 2002, 2003, 2004 Joel Uckelman
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "game.h"
#include "view.h"
#include "surface.h"


//
// Surface
//
#ifdef Q_WS_WIN
//
// Windows chokes on QPixmaps larger than 2000x2000 or so.
// That means we can't set the surface background as a true 
// background if it's too large. Argh!
//
#include <math.h>  // for ceil()

#include <qvaluelist.h>

Surface::Surface(GameObj *g, QCanvas *c, const QString &n, const QFileInfo &b,
 const QImage p, unsigned int i, QWidget *parent, const char *name, WFlags f) :
 View(g, c, parent, name, f), Name(n), bgfile(b), ID(i ? i : g->maxID++) {
   // set background
   if (p.isNull()) canvas()->resize(300, 500);
   else {
      const unsigned int tsize = 2000;
      const unsigned int width = p.width();
      const unsigned int height = p.width();
      const unsigned int xtiles = (unsigned int)ceil((double)width / tsize);
      const unsigned int ytiles = (unsigned int)ceil((double)height / tsize);

      // chop the background into tiles
      for (unsigned int x = 0; x < xtiles; ++x) {
         for (unsigned int y = 0; y < ytiles; ++y) {
            const unsigned int w =
             width > (x+1)*tsize ? tsize : width-(x*tsize);
            const unsigned int h =
             height > (y+1)*tsize ? tsize : height-(y*tsize);

            QValueList<QPixmap> pl;
            QPixmap pm(p.copy(x*tsize, y*tsize, w, h));
            pl.append(pm);
            QCanvasPixmapArray *pa = new QCanvasPixmapArray(pl);

            QCanvasSprite *tile = new QCanvasSprite(pa, c);
            tile->setX(x*tsize);
            tile->setY(y*tsize);
            tile->setZ(-1);
            tile->show();            
         }
      }

      canvas()->resize(p.width(), p.height());
      //setMaximumSize(sizeHint());
   }

   setName(Name.ascii());
   if (i >= g->maxID) g->maxID = i + 1;
}
#else
//
// This is the Surface constructor for all sane platforms.
//
Surface::Surface(GameObj *g, QCanvas *c, const QString &n, const QFileInfo &b,
 const QPixmap p, unsigned int i, QWidget *parent, const char *name, WFlags f)
 : View(g, c, parent, name, f), Name(n), bgfile(b), ID(i ? i : g->maxID++) {
   // set background
   if (p.isNull()) canvas()->resize(300, 500);
   else {
      canvas()->setBackgroundPixmap(p);
      canvas()->resize(p.width(), p.height());
      // FIXME: do size hinting better
      //setMaximumSize(sizeHint());
   }

   setName(Name.ascii());
   if (i >= g->maxID) g->maxID = i + 1;
}
#endif


void Surface::setSurfaceName(const QString &n) { 
   Name = n;
   setName(Name.ascii());
}


QString Surface::surfaceName() const { return Name; }


QFileInfo Surface::backFile() const { return bgfile; }


void Surface::setBackFile(const QFileInfo &bf) { bgfile = bf; }


void Surface::showEvent(QShowEvent *) { emit shown(this); }


void Surface::hideEvent(QHideEvent *) { emit hidden(this); }


void Surface::mousePressEvent(QMouseEvent *) { emit pressed(this); }
        

void Surface::resizeEvent(QResizeEvent *e) {
   if (bgfile.filePath().isEmpty()) {
      if (canvas()->width() < e->size().width() || 
       canvas()->height() < e->size().height()) {
         int w = QMAX(canvas()->width(), e->size().width());
         int h = QMAX(canvas()->height(), e->size().height());
         canvas()->resize(w, h);
      }
   }
   
   View::resizeEvent(e);
}


unsigned int Surface::id() const { return ID; }


QDomElement Surface::dumpXML(QDomDocument &doc) const {
   QDomElement e = doc.createElement("surface");
   e.setAttribute("name", Name);
   e.setAttribute("id", id());
   if (!bgfile.filePath().isEmpty())
    e.setAttribute("background", bgfile.fileName());
   return e;
}

