// $Id: surfacedialogs.cpp,v 1.3 2004/07/29 16:47:25 uckelman Exp $
/* Board -- a graphical frontend for boardgames
 * Copyright (C) 2003, 2004 Joel Uckelman
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <qcanvas.h>
#include <qheader.h>
#include <qinputdialog.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qtooltip.h>

#include "action.h"
#include "game.h"
#include "surfacedialogs.h"

#include "tools/clone.xpm"
#include "tools/new.xpm"
#include "tools/trash.xpm"

#include "icons/maps.xpm"

#include <qheader.h>

//
// SurfaceView
//
SurfaceView::SurfaceView(GameObj *g, QWidget *parent, const char *name) :
 EditView(parent, name), game(g) {
   addColumn("Name");
   header()->hide();
}


void SurfaceView::rename() {
   QString et = edit->text();
   if (!et.isEmpty()) {
      QString n = renameItem->text(0);
      if (et != n) {
         Surface *s =
          ((SurfaceListDialog *)parent())->s2i[(QCheckListItem *)renameItem];
         Action *a = new SurfaceRenameAction(s, et);
         if (a->isPermissible()) {
            a->commit();
            game->done.log(a);
            game->done.complete();
         }
         else delete a;
      }
   }
   cancelRename();
}


//
// SurfaceViewCheckItem
//
SurfaceViewCheckItem::SurfaceViewCheckItem(SurfaceView *parent,
 const QString &text) : QCheckListItem(parent, text, CheckBox) {
   if (parent) parent->hasCheckBox.append(this);
}


SurfaceViewCheckItem::~SurfaceViewCheckItem() {
   if (listView()) ((SurfaceView *)listView())->hasCheckBox.remove(this);
}


void SurfaceViewCheckItem::activate() {
   bool on = isOn();
   QCheckListItem::activate();
   if (on != isOn()) emit ((SurfaceView *)listView())->toggled(this);
}


//
// SurfaceListDialog
//
SurfaceListDialog::SurfaceListDialog(GameObj *g, QWidget *parent,
 const char *name, WFlags f) :
 TabPage(QIconSet(QPixmap((const char **)maps_xpm)),
 "Surfaces", parent, name, f), game(g) {
   QVBoxLayout *vl = (QVBoxLayout *)layout();

   // view
   view = new SurfaceView(game, this);
   connect(view, SIGNAL(toggled(QCheckListItem *)),
    this, SLOT(toggleVisibility(QCheckListItem *)));
   connect(view, SIGNAL(selectionChanged(QListViewItem *)),
    this, SLOT(raiseActive(QListViewItem *)));
   connect(view, SIGNAL(selectionChanged(QListViewItem *)),
    this, SLOT(changedSelection(QListViewItem *)));
   vl->addWidget(view);

   // buttons
   QHBoxLayout *bl = new QHBoxLayout(vl);
   
   // new
   nb = new QPushButton(this);
   nb->setPixmap(QPixmap((const char **)new_xpm));
   nb->setFixedSize(27, 27);
   QToolTip::add(nb, "New");
   connect(nb, SIGNAL(clicked()), this, SLOT(createSurface()));
   bl->addWidget(nb);

   // trash 
   tb = new QPushButton(this);
   tb->setPixmap(QPixmap((const char **)trash_xpm));
   tb->setFixedSize(27, 27);
   QToolTip::add(tb, "Delete");
   connect(tb, SIGNAL(clicked()), this, SLOT(deleteSurface()));
   bl->addWidget(tb);

   // clone 
   cb = new QPushButton(this);
   cb->setPixmap(QPixmap((const char **)clone_xpm));
   cb->setFixedSize(27, 27);
   QToolTip::add(cb, "Clone");
   connect(cb, SIGNAL(clicked()), this, SLOT(copySurface()));
   bl->addWidget(cb);

   bl->addStretch();
}


void SurfaceListDialog::insert(Surface *s) {
   QCheckListItem *i = new SurfaceViewCheckItem(view, s->name());
   s2i.insert(s, i);
   connect(s, SIGNAL(pressed(Surface *)), this, SLOT(setActive(Surface *)));
   connect(s, SIGNAL(shown(Surface *)), this,
    SLOT(toggleVisibility(Surface *)));
   connect(s, SIGNAL(hidden(Surface *)), this,
    SLOT(toggleVisibility(Surface *)));
}


void SurfaceListDialog::remove(Surface *s) {
   delete s2i[s];
   s2i.remove(s);
}


void SurfaceListDialog::rename(Surface *s) {
   QListViewItem *i = s2i[s];
   i->setText(0, s->surfaceName());
}


void SurfaceListDialog::setActive(Surface *s) {
   QListViewItem *i = s2i[s];
   view->setCurrentItem(i);
}


void SurfaceListDialog::toggleVisibility(Surface *s) {
   if (s) {
      QCheckListItem *i = s2i[s];
      if (i) i->setOn(s->isVisible());
   } 
}


void SurfaceListDialog::toggleVisibility(QCheckListItem *i) {
   if (i) {
      Surface *s = s2i[i];
      if (s) {
         // FIXME: surface items should be toggled off when surfaces are
         // minimized; there's no obvious way to do this in Qt 3.2.1
         // without reimplementing QWidget::event().
         if (i->isOn()) s->show();
         else s->hide();
      }
   }
}


void SurfaceListDialog::raiseActive(QListViewItem *i) {
   Surface *s = s2i[(QCheckListItem *)i];
   if (s->isVisible()) {
      s->setActiveWindow();
      s->raise();
   }
}


void SurfaceListDialog::createSurface() {
   // get surface name
   bool ok = false;
   QString name = QInputDialog::getText("Create a Surface", "Surface Name:",
    QLineEdit::Normal, QString::null, &ok, this);
   if (!ok) return;

   // take action
   SurfaceCreateAction *a = new SurfaceCreateAction(name);
   if (a->isPermissible()) {
      a->commit();
      game->done.log(a);
      game->done.complete();
   }
   else delete a;
}


void SurfaceListDialog::deleteSurface() {
   QCheckListItem *i = (QCheckListItem *)view->selectedItem();
   if (i) {
      Surface *s = s2i[i];
      SurfaceTrashAction *a = new SurfaceTrashAction(s);
      
      if (a->isPermissible()) {
         a->commit();
         game->done.log(a);
         game->done.complete();
      }
      else delete a;
   }
}


void SurfaceListDialog::copySurface() {
   // FIXME: reinstate this once SurfaceCloneAction is converted
   /*
   QCheckListItem *i = (QCheckListItem *)view->selectedItem();
   if (i) {
      Surface *s = s2i[i];
      SurfaceCloneAction *a = new SurfaceCloneAction(s);

      if (a->isPermissible()) {
         a->commit();
         game->done.log(a);
         game->done.complete();
      }
      else delete a;
   } 
   */
}


void SurfaceListDialog::changedSelection(QListViewItem *i) {
   Surface *s = s2i[(QCheckListItem *)i];
   QValueList<Token *> l = s->tokens();
   bool p = true;
   for (QValueList<Token *>::const_iterator i = l.begin(); i != l.end() && p;
    ++i) p &= (*i)->permission(game->user, Perms::Delete);

   // disable trash if no selection or no permission
   tb->setEnabled(s && p);
   
   // disable copy if no selection
   cb->setEnabled(s);

   // FIXME: buttons should be updated when permissions change as well
}


