// $Id: tabdialog.cpp,v 1.17 2004/08/02 22:52:55 uckelman Exp $
/* Board -- a graphical frontend for boardgames
 * Copyright (C) 2003, 2004 Joel Uckelman
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <iostream>

using namespace std;

#include <qapplication.h>
#include <qdatastream.h>
#include <qlayout.h>
#include <qtabbar.h>

#include "tabdialog.h"
#include "tabpage.h"

extern QString TabMIMEtype;


//
// TabDrag
//
TabDrag::TabDrag(TabPage *page, QWidget *dragSource, const char *name) :
 QStoredDrag(TabMIMEtype, dragSource, name) {
   QByteArray bytes;
   QDataStream str(bytes, IO_WriteOnly);
   str << (Q_ULONG)page;
   setEncodedData(bytes);
}


bool TabDrag::canDecode(const QMimeSource *m) {
   return m->provides(TabMIMEtype);
}


bool TabDrag::decode(const QMimeSource *m, TabPage *&page) {
   QDataStream str(m->encodedData(TabMIMEtype), IO_ReadOnly);
   Q_ULONG ptr;
   str >> ptr;
   page = (TabPage *)ptr;
   return true;
}


//
// TabDialog
//
TabDialog::TabDialog(QWidget *parent, const char *name, WFlags f) :
 QWidget(parent, name, f) {
   QVBoxLayout *vb = new QVBoxLayout(this, 2);

   tw = new TabHolder(this);
   vb->addWidget(tw);

   setAcceptDrops(true);
}


TabDialog::~TabDialog() {
   // set children free, whatever put them here should have
   // retained a pointer 
   tw->setCurrentPage(0);
   QWidget *w;

   while ((w = tw->currentPage())) w->reparent(0, QPoint(0,0));
}


void TabDialog::addPage(TabPage *page) {
   tw->addPage(page);
}


void TabDialog::dragEnterEvent(QDragEnterEvent *e) {
   tw->dragEnterEvent(e);
}


void TabDialog::dropEvent(QDropEvent *e) {
   tw->dropEvent(e);
}


//
// TabHolder
//
TabHolder::TabHolder(QWidget *parent, const char *name) :
 QTabWidget(parent, name) {
   setTabBar(new TabBar(this));
   setAcceptDrops(true);

   // The page stack is private to QTabWidget, so in order to flatten
   // its frame, we have to call QTabWidget::removePage(), which
   // flattens the stack frame if the page stack is empty. Ugly.
   QWidget *w = new QWidget;
   QTabWidget::addTab(w, QString::null);
   QTabWidget::removePage(w);
   delete w;
}

void TabHolder::addPage(TabPage *page) {
   // add the page
   QTabWidget::addTab(page, page->iconSet(), QString::null);
   setTabToolTip(page, page->toolTip());

   connect(page, SIGNAL(requestClose(TabPage *)),
    this, SLOT(removePage(TabPage *)));

   // hide tabs if this is a single page
   // show tabs if there are now multiple pages
   switch (count()) {
      case 0:
      case 1: tabBar()->hide(); break;
      default: tabBar()->show(); break;
   }
}


void TabHolder::removePage(TabPage *page, bool autoClose) {
   // remove the page
   QTabWidget::removePage(page);
   disconnect(page, SIGNAL(requestClose(TabPage *)),
    this, SLOT(removePage(TabPage *)));
   page->reparent(0, QPoint(0,0));

   // hide tabs if this would leave only one page
   // maybe close parent if holder is empty
   switch (count()) {
      case 0: 
         if (autoClose && dynamic_cast<TabDialog *>(parentWidget())) {
            parentWidget()->close();
            break;
         }
      case 1: tabBar()->hide(); break;      
      default: break;
   }
}



void TabHolder::dragEnterEvent(QDragEnterEvent *e) {
   e->accept(TabDrag::canDecode(e));
}


void TabHolder::dropEvent(QDropEvent *e) {
   if (!TabDrag::canDecode(e)) {
      e->ignore();
      return;
   }

   TabPage *page;
   if (TabDrag::decode(e, page)) {
      TabHolder *src = (TabHolder *)e->source();
      QIconSet i = src->tabIconSet(page);
      QString t = src->tabToolTip(page);
      src->removePage(page, (src != this) );

      addPage(page);
      showPage(page);

      e->accept();
   }
   else e->ignore();
}


void TabHolder::mouseMoveEvent(QMouseEvent *e) {
   TabPage *current = (TabPage *)currentPage();

   // nothing to do if there are no pages
   if (!current) return;

   // start drag only if cursor has moved far enough since button press
   if (pressed && (presspos - e->pos()).manhattanLength() >
    QApplication::startDragDistance()) {
      pressed = false;
      TabDrag *td = new TabDrag(current, this);
      td->drag();
   }
}


void TabHolder::mousePressEvent(QMouseEvent *e) {
   presspos = e->pos();
   pressed = true;
}


void TabHolder::mouseReleaseEvent(QMouseEvent *) { 
   pressed = false;
}


//
// TabBar
//
TabBar::TabBar(QWidget *parent, const char *name) : QTabBar(parent, name) {
   setAcceptDrops(true);
}


void TabBar::dragEnterEvent(QDragEnterEvent *e) {
   e->accept(TabDrag::canDecode(e));
}


void TabBar::dropEvent(QDropEvent *e) {
   ((TabHolder *)parentWidget())->dropEvent(e);
}


void TabBar::mouseMoveEvent(QMouseEvent *e) {
   QTabBar::mouseMoveEvent(e);   
   ((TabHolder *)parentWidget())->mouseMoveEvent(e);
}


void TabBar::mousePressEvent(QMouseEvent *e) {
   QTabBar::mousePressEvent(e);
   ((TabHolder *)parentWidget())->mousePressEvent(e);
}


void TabBar::mouseReleaseEvent(QMouseEvent *e) { 
   QTabBar::mouseReleaseEvent(e);
   ((TabHolder *)parentWidget())->mouseReleaseEvent(e);
}
