// $Id: token.cpp,v 1.14 2004/10/21 21:17:23 uckelman Exp $
/* Board -- a graphical frontend for boardgames
 * Copyright (C) 2002, 2003, 2004 Joel Uckelman
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <qbitmap.h>
#include <qcanvas.h>
#include <qmap.h>
#include <qpainter.h>

// VC++ needs this for M_PI
#ifdef Q_CC_MSVC
#define _USE_MATH_DEFINES
#endif

#include <math.h>

#include "game.h"
#include "token.h"


//
// Token
// 
Token::Token(GameObj *g, Group *gr, QCanvas *canvas, int x, int y, int z,
 int a, const QString &n, unsigned int t, unsigned int u, unsigned int i) :
 QCanvasPolygon(canvas), Groupable(g, n, gr, i), Type(t), Theta(a),
 selection(6) {
   // set showing face
   // FIXME: check whether tTypes contains Type
   pl = g->tTypes[Type]->side;
   for (side = pl.begin(); side != pl.end() && (*side).id() != u; ++side);

   adjustCorners();
    
   // set location
   setX(x);
   setY(y);
   setZ(z);
}


Token::Token(const Token &c, unsigned int i) :  QCanvasPolygon(c.canvas()),
 Groupable(c.game, c.Name, c.Parent, i), pl(c.pl), Type(c.Type),
 Theta(c.Theta), m(c.m), r(c.r), selection(6) {
   
   //set showing face
   for (side = pl.begin(); side != pl.end() && (*side).id() != c.up(); ++side);

   // set corners
   setPoints(c.points());

   // set location
   setX(c.x());
   setY(c.y());
   setZ(c.z()+1);

   // set selected
   if (c.selected()) setSelected(true);
}


Token::~Token() {
   hide();
}


void Token::show() { QCanvasPolygon::show(); }


void Token::hide() { QCanvasPolygon::hide(); }


bool Token::isVisible() const { return QCanvasPolygon::isVisible(); }

void Token::drawShape(QPainter &p) {
   QRect s = (*side).rect();
   QPoint v((r.width()-s.width())/2, (r.height()-s.height())/2);   

   p.setRasterOp(Qt::CopyROP);
   p.drawPixmap(int(x()-v.x()), int(y()-v.y()), r);
  
   if (selected()) {
      p.setPen(QPen("white", 0, Qt::DotLine));
      p.setRasterOp(Qt::XorROP);
      p.drawPolyline(selection);
   }
  
   /* 
   // draw item edge, bounding rectangle; useful for debugging
   p.setPen(QColor("red"));
   p.drawRect(boundingRect());
   p.setPen(QColor("blue"));
   p.drawPolyline(areaPoints());
   */
}


int Token::theta() const { return Theta; }


void Token::setTheta(int t) {
   if (Theta == t) return;
   Theta = t % 360;
   
   adjustCorners();

   // transform selection
   if (selected()) {
      selection = m.map(selection);
   }
           
   canvas()->setChanged(boundingRect()); 
}


QWMatrix Token::transMatrix(double x, double y, double t) const {
   //
   // transformation matrix for rotating an object about (x,y):
   //  
   //         cos(t)                     sin(t)                0
   //        -sin(t)                     cos(t)                0
   //  -x*cos(t)+y*sin(t)+y       -x*sin(t)-y*cos(t)+y         1
   //        
   double rad = M_PI*t/180;
   double c = cos(rad);
   double s = sin(rad);
   return QWMatrix(c, s, -s, c, -x*c+y*s+x, -x*s-y*c+y);
}


void Token::adjustCorners() {
   // reset transformation matrix
   QRect s = (*side).rect();
   m = transMatrix(s.width()/2, s.height()/2, Theta);

   // transform pixmap
   r = (*side).xForm(m);
   if (Theta % 90) r.setMask(r.createHeuristicMask());

   // reset corners
   setPoints(m.map(QPointArray(QRect(s.topLeft()-QPoint(1,1),
    s.bottomRight()+QPoint(1,1)), true)));
}
        

int Token::rtti() const { return 500000; }


void Token::setX(double x) {
   selection.translate(int(x - selection[0].x()), 0);
   QCanvasPolygon::setX(x);
}


void Token::setY(double y) {
   selection.translate(0, int(y - selection[0].y()));
   QCanvasPolygon::setY(y);
}


void Token::moveBy(double dx, double dy) {
   selection.translate(int(dx), int(dy));
   QCanvasPolygon::moveBy(dx, dy);
}


void Token::setSelected(bool selected) {
   if (selected) {
      // FIXME: does not rotate when token is rotated
      // set selection
      QRect b = boundingRect();
      selection[0] = b.topLeft();
      selection[1] = b.topRight();
      selection[2] = b.bottomRight();
      selection[3] = b.bottomLeft();
      selection[4] = selection[0];
      selection[5] = selection[0];      
   }

   QCanvasItem::setSelected(selected);
}


void Token::advanceAnts(unsigned int a) {
   selection[0] = selection[5] =
    QPoint(selection[3].x()+a, selection[0].y());
   canvas()->setChanged(boundingRect());
}
        

unsigned int Token::type() const { return Type; }


unsigned int Token::up() const { return (*side).id(); }


QPixmap Token::face() const { return *side; }


void Token::flipNext() {
   if ((++side) == pl.end()) side = pl.begin();  
 
   // transform pixmap
   r = (*side).xForm(m);
   if (Theta % 90) r.setMask(r.createHeuristicMask());
   canvas()->setChanged(boundingRect()); 
}


void Token::flipPrev() {
   if (side == pl.begin()) side = pl.end();
   --side;

   // transform pixmap
   r = (*side).xForm(m);
   if (Theta % 90) r.setMask(r.createHeuristicMask());
   canvas()->setChanged(boundingRect()); 
}


bool Token::permission(const Player *pl, unsigned int pm) const {
   return Parent ? Parent->permission(pl, pm) : true;
}


QDomElement Token::dumpXML(QDomDocument &doc) const {
   QDomElement e = doc.createElement("token");
   e.setAttribute("id", id());
   e.setAttribute("x", x());
   e.setAttribute("y", y());
   e.setAttribute("z", z());
   e.setAttribute("t", Theta);
   e.setAttribute("type", Type);
   e.setAttribute("up", up()); 
   e.setAttribute("name", Name);
   e.setAttribute("surface", game->c2s[canvas()]->id());
   return e;
}

/*
QTextStream &Token::operator<<(QTextStream &s) {
   s << "<token id=\"" << id() << "\" x=\"" << x() << "\" y=\"" << y() <<
    "\" z=\"" << z() << "\" t=\"" << Theta << "\" type=\"" << Type <<
    "\" up=\"" << up() << "\" name=\"" << Name << "\" tray=\"" <<
    game->c2s[canvas()]->id() << "\" />\n";
   return s;
}
*/

//
// TokenPixmap
//
TokenPixmap::TokenPixmap() { }


TokenPixmap::TokenPixmap(unsigned int id) : pixid(id) { }


TokenPixmap::TokenPixmap(const QPixmap &pix, unsigned int id) :
 QPixmap(pix), pixid(id) { }

 
unsigned int TokenPixmap::id() const { return pixid; }
