// $Id: tokendialogs.cpp,v 1.8 2004/08/03 22:02:06 uckelman Exp $
/* Board -- a graphical frontend for boardgames
 * Copyright (C) 2003, 2004 Joel Uckelman
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <qlabel.h>
#include <qlayout.h>
#include <qlistbox.h>
#include <qmenudata.h>
#include <qpainter.h>

#include "action.h"
#include "game.h"
#include "groupdialogs.h"
#include "player.h"
#include "token.h"
#include "tokendialogs.h"
#include "tokendrag.h"

#include "icons/tokens.xpm"

extern GameObj *Game;

//
// TokenListBoxItem
//
TokenListBoxItem::TokenListBoxItem(TokenType *d, QListBox *listbox) :
 QListBoxItem(listbox), def(d) {
   setText(def->name);
}


int TokenListBoxItem::width(const QListBox *lb) const {
   int w = 0;
   for (TokenPixmapList::const_iterator i = def->side.begin();
    i != def->side.end(); ++i) w += (*i).width() + 2;
 
   if (!text().isEmpty()) w += lb->fontMetrics().width(text()) + 12;
   
   return w;
}


int TokenListBoxItem::height(const QListBox *lb) const {
   int h = 0;
   for (TokenPixmapList::const_iterator i = def->side.begin();
    i != def->side.end(); ++i) h = QMAX((*i).height(), h);

   return QMAX(lb->fontMetrics().lineSpacing() + 1, h + 2);
}


TokenType *TokenListBoxItem::definition() const { return def; }


void TokenListBoxItem::paint(QPainter *painter) {
   int x = 3;
   
   for (TokenPixmapList::const_iterator i = def->side.begin();
    i != def->side.end(); x += (*i).width()+2, ++i)
    painter->drawPixmap(x, 1, *i);

   if (!text().isEmpty()) {
      QFontMetrics fm = painter->fontMetrics();
      int yPos;
      
      if (height(listBox()) < fm.height()) yPos = fm.ascent() + fm.leading()/2;
      else yPos = height(listBox())/2 - fm.height()/2 + fm.ascent();
      
      painter->drawText(x + 5, yPos, text());     
   }
}


//
// TokenListBox
//
TokenListBox::TokenListBox(QWidget *parent, const char *name, WFlags f) :
 QListBox(parent, name, f), moving(0) {
}


void TokenListBox::append(TokenListBoxItem *i) { insertItem(i); }


void TokenListBox::viewportMousePressEvent(QMouseEvent *e) {
   if (e->button() == LeftButton) moving = (TokenListBoxItem*)itemAt(e->pos());
   else moving = 0;
}


void TokenListBox::viewportMouseMoveEvent(QMouseEvent *) {
   if (moving) {
      TokenCreateAction *create = new TokenCreateAction(moving->definition(),
       Game->groupsdialog->currentGroup());
      if (create->isPermissible()) {
         create->begin();
         Game->done.log(create);
      
         QDragObject *d = new TokenDrag(create->created(), QPoint(0,0), this);
         connect(d, SIGNAL(failed()), &Game->done, SLOT(abort()));
         d->dragMove();      
      }
      else delete create;
      
      moving = 0;
   }
}


//
// TokenListDialog
//
TokenListDialog::TokenListDialog(QWidget *parent, const char *name, WFlags f) :
 TabPage(QIconSet(QPixmap((const char **)tokens_xpm)), "Token List", parent,
 name, f) {
   QVBoxLayout *vl = (QVBoxLayout *)layout();

   // list box
   listBox = new TokenListBox(this);
   vl->addWidget(listBox);
}


void TokenListDialog::append(TokenType *type) {
   listBox->append(new TokenListBoxItem(type));
}


//
// TokenListView
//
TokenListView::TokenListView(GameObj *g, QWidget *parent, const char *name) :
 QListView(parent, name), game(g) {
   addColumn("Token");
   addColumn("Name");
   addColumn("Map");
   addColumn("Location");
   addColumn("Group");
   addColumn("Owner");

   setShowSortIndicator(true);
   setAllColumnsShowFocus(true);
}


void TokenListView::insert(Token *t) {
   if (t2i.contains(t)) return;

   QString loc;
   loc.sprintf("(%.0f,%.0f)", t->x(), t->y());
   
   Player *p = t->owner();
   QString owner = p ? p->name() : QString::null;
     
   QListViewItem *i = new QListViewItem(this, QString::null, t->name(),
    game->c2s[t->canvas()]->surfaceName(), loc, t->parent()->name(), owner);

   i->setPixmap(0, t->face());

   t2i.insert(t, i);
}


void TokenListView::remove(Token *t) {
   delete t2i[t];
   t2i.remove(t);
}


void TokenListView::update(unsigned int column, Token *t) {
   if (!t2i.contains(t)) return;
   QListViewItem *i = t2i[t];

   QString s;
   switch (column) {
      case 0:
         i->setPixmap(0, t->face());
         break;
      case 1:
         i->setText(1, t->name());
         break;
      case 2:
         i->setText(2, game->c2s[t->canvas()]->surfaceName());
         break;
      case 3:
         s.sprintf("(%.0f,%.0f)", t->x(), t->y());
         i->setText(3, s);
         break;
      case 4:
         i->setText(4, t->parent()->name());
         break; 
      case 5:
         i->setText(5, t->owner() ? t->owner()->name() : QString::null);
         break;
      default:
         break;
   }
}


void TokenListView::clear() {
   QListViewItem *i;
   while((i = firstChild())) remove(t2i[i]);
}


//
// TokenListViewDialog
//
TokenListViewDialog::TokenListViewDialog(GameObj *g, const QIconSet &icon,
 const QString &title, QWidget *parent, const char *name, WFlags f) :
 TabPage(icon, title, parent, name, f) {
   QVBoxLayout *vl = (QVBoxLayout *)layout();

   // view
   view = new TokenListView(g, this);
   vl->addWidget(view);
}


void TokenListViewDialog::insert(Token *c) {
   view->insert(c);
}


void TokenListViewDialog::remove(Token *c) {
   view->remove(c);
}


void TokenListViewDialog::update(unsigned int column, Token *c) {
   view->update(column, c);
}


void TokenListViewDialog::clear() {
   view->clear();
}
