// $Id: view.cpp,v 1.134 2004/09/21 15:13:09 uckelman Exp $
/* Board -- a graphical frontend for boardgames
 * Copyright (C) 2002, 2003, 2004 Joel Uckelman
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <iostream>
using namespace std;

#include <math.h>

#include <qaccel.h>
#include <qapplication.h>
#include <qbitmap.h>
#include <qcanvas.h>
#include <qclipboard.h>
#include <qcursor.h>
#include <qevent.h>
#include <qpainter.h>
#include <qpopupmenu.h>
#include <qtimer.h>

#include "action.h"
#include "freehand.h"
#include "game.h"
#include "groupdialogs.h"
#include "main.h"
#include "player.h"
#include "polyline.h"
#include "preferences.h"
#include "properties.h"
#include "selection.h"
#include "selectionhandler.h"
#include "token.h"
#include "tokendialogs.h"
#include "tokendrag.h"
#include "view.h"

#include "cursors/std.xbm"
#include "cursors/std_mask.xbm"
#include "cursors/std_plus.xbm"
#include "cursors/std_plus_mask.xbm"
#include "cursors/std_minus.xbm"
#include "cursors/std_minus_mask.xbm"
#include "cursors/std_union.xbm"
#include "cursors/std_union_mask.xbm"
#include "cursors/std_inter.xbm"
#include "cursors/std_inter_mask.xbm"

extern PreferenceMap Pref;
extern Main *MW;

const int ScrollMargin = 16;
const int ScrollTimeInitial = 30;


//
// View
//
const QCursor *View::NormalCursor = 0;
const QCursor *View::PlusCursor = 0;
const QCursor *View::MinusCursor = 0;
const QCursor *View::UnionCursor = 0;
const QCursor *View::IntersectCursor = 0;

QMap<QString, char *> View::Slot;


View::View(GameObj *g, QCanvas *c, QWidget *parent, const char *name,
 WFlags f) : QCanvasView(c, parent, name, f), game(g), raised(0), selecting(0),
 freehand(0), plot(0), pline(0), scroll_timer(this), compound(false), Zoom(1),
 menu(this), accel(this) {
 
   // set function pointers for tools
   toolFunc[TMove] = &View::toolMove;
   toolFunc[TClone] = &View::toolClone;
   toolFunc[TPlot] = &View::toolPlot;
   toolFunc[TTrash] = &View::toolTrash;
   toolFunc[TFlipLeft] = &View::toolFlipLeft;
   toolFunc[TFlipRight] = &View::toolFlipRight;
   toolFunc[TRotateCW] = &View::toolRotateCW;
   toolFunc[TRotateCCW] = &View::toolRotateCCW;
   toolFunc[TRegroup] = &View::toolRegroup;
   toolFunc[TSelect] = &View::toolSelect;
   toolFunc[TWand] = &View::toolWand;
   toolFunc[TCrayon] = &View::toolCrayon;
   toolFunc[TZoomIn] = &View::toolZoomIn;
   toolFunc[TZoomOut] = &View::toolZoomOut;
   toolFunc[TProp] = &View::toolProp;
   toolFunc[TRaise] = &View::toolRaise;
   
   // popup menu for view-specific functions
   menu.insertItem("&Undo", MW, SLOT(editUndo()), CTRL+Key_Z, mUndo);
   menu.insertItem("&Redo", MW, SLOT(editRedo()), CTRL+Key_R, mRedo);
   menu.insertSeparator();
   menu.insertItem("&Cut", this, SLOT(cut()), CTRL+Key_X, mCut);
   menu.insertItem("C&opy", this, SLOT(copy()), CTRL+Key_C, mCopy);
   menu.insertItem("&Paste", this, SLOT(paste()), CTRL+Key_V, mPaste);
   menu.insertSeparator();
   menu.insertItem("Invert", this, SLOT(selectInvert()), CTRL+Key_I,
    mSelectInvert);
   menu.insertItem("All", this, SLOT(selectAll()), CTRL+Key_A,
    mSelectAll);
   menu.insertItem("None", this, SLOT(selectNone()), SHIFT+CTRL+Key_A,
    mSelectNone);
   menu.insertSeparator();
   menu.insertItem("Offset Stack", this, SLOT(stackOffset()), CTRL+Key_H,
    mStackOffset);
   menu.insertItem("Flush Stack", this, SLOT(stackFlush()), 0,
    mStackFlush);
   menu.insertItem("Shuffle", this, SLOT(stackShuffle()), CTRL+Key_J,
    mStackShuffle);
   menu.insertItem("Invert Stack", this, SLOT(stackInvert()), 0,
    mStackInvert);
   menu.insertSeparator();
   menu.insertItem("Properties", this, SLOT(properties()), 0, mProperties);
   
   // create accel group
   accel.connectItem(accel.insertItem(CTRL+Key_N, M_FILE_NEW),
    MW, SLOT(fileNew()));
   accel.connectItem(accel.insertItem(CTRL+Key_O, M_FILE_OPEN),
    MW, SLOT(fileOpen()));
   accel.connectItem(accel.insertItem(CTRL+Key_S, M_FILE_SAVE),
    MW, SLOT(fileSave()));
   accel.connectItem(accel.insertItem(CTRL+Key_W, M_FILE_CLOSE),
    MW, SLOT(fileClose()));
   accel.connectItem(accel.insertItem(CTRL+Key_Q, M_FILE_QUIT),
    MW, SLOT(close()));
   accel.connectItem(accel.insertItem(Key_F1, M_HELP_HELP),
    MW, SLOT(helpHelp()));

   // arrow keys for scrolling
   accel.connectItem(accel.insertItem(Key_Up), this, SLOT(scrollUp()));
   accel.connectItem(accel.insertItem(Key_Down), this, SLOT(scrollDown()));
   accel.connectItem(accel.insertItem(Key_Left), this, SLOT(scrollLeft()));
   accel.connectItem(accel.insertItem(Key_Right), this, SLOT(scrollRight()));

   // populate Slot if not done already
   if (Slot.isEmpty()) {
      Slot["key.tool.move"] = SLOT(setToolMove());
      Slot["key.tool.plot"] = SLOT(setToolPlot());
      Slot["key.tool.select"] = SLOT(setToolSelect());
      Slot["key.tool.wand"] = SLOT(setToolWand());
      Slot["key.tool.clone"] = SLOT(setToolClone());
      Slot["key.tool.trash"] = SLOT(setToolTrash());
      Slot["key.tool.flip_left"] = SLOT(setToolFlipLeft());
      Slot["key.tool.flip_right"] = SLOT(setToolFlipRight());
      Slot["key.tool.rotate_ccw"] = SLOT(setToolRotateCCW());
      Slot["key.tool.rotate_cw"] = SLOT(setToolRotateCW());
      Slot["key.tool.regroup"] = SLOT(setToolRegroup());
      Slot["key.tool.zoom_in"] = SLOT(setToolZoomIn());
      Slot["key.tool.zoom_out"] = SLOT(setToolZoomOut());
      Slot["key.tool.properties"] = SLOT(setToolProperties());
   }
 
   // build accellerators for key bindings
   // and callbacks to catch binding changes
   for (QMap<QString, char *>::const_iterator i = Slot.begin();
    i != Slot.end(); ++i) {
      int key = Pref[i.key()].effective().toInt();
      binding2key[i.key()] = key;
      accel.insertItem(key, key);
      accel.connectItem(key, MW, i.data());

      connect(MW, SIGNAL(preferenceChanged(const QString)),
       this, SLOT(rebindKey(const QString)));
   }
   
   // enable, disable edit options as appropriate
   setSelected(false);
   setPasteable(TokenDrag::canDecode(QApplication::clipboard()->data())
    ? true : false);

   // create NormalCursor if not there already
   if (!NormalCursor) NormalCursor = new QCursor(
    QBitmap(std_width, std_height, std_bits, true),
    QBitmap(std_mask_width, std_mask_height, std_mask_bits, true),
    std_x_hot, std_y_hot);
  
   // create PlusCursor if not there already
   if (!PlusCursor) PlusCursor = new QCursor(
    QBitmap(std_plus_width, std_plus_height, std_plus_bits, true),
    QBitmap(std_plus_mask_width, std_plus_mask_height, std_plus_mask_bits,
     true), std_plus_x_hot, std_plus_y_hot);

   // create MinusCursor if not there already
   if (!MinusCursor) MinusCursor = new QCursor(
    QBitmap(std_minus_width, std_minus_height, std_minus_bits, true),
    QBitmap(std_minus_mask_width, std_minus_mask_height, std_minus_mask_bits,
     true), std_minus_x_hot, std_minus_y_hot);
      
   // create UnionCursor if not there already
   if (!UnionCursor) UnionCursor = new QCursor(
    QBitmap(std_union_width, std_union_height, std_union_bits, true),
    QBitmap(std_union_mask_width, std_union_mask_height, std_union_mask_bits,
     true), std_union_x_hot, std_union_y_hot);
   
   // create IntersectCursor if not there already
   if (!IntersectCursor) IntersectCursor = new QCursor(
    QBitmap(std_inter_width, std_inter_height, std_inter_bits, true),
    QBitmap(std_inter_mask_width, std_inter_mask_height, std_inter_mask_bits,
     true), std_inter_x_hot, std_inter_y_hot);
   
   // normal cursor is default
   setCursor(*NormalCursor);

   // connect timers
   connect(&scroll_timer, SIGNAL(timeout()), this, SLOT(doAutoScroll()));

   // connect to selection handler
   connect(game->selector, SIGNAL(hasSelection(bool)),
    this, SLOT(setSelected(bool)));

   // DnD stuff
   viewport()->setAcceptDrops(true);
   setDragAutoScroll(false);
}
  

View::~View() {
}


QString View::name() const { return QString(QCanvasView::name()); }


void View::setName(const QString &n) { QCanvasView::setName(n.latin1()); }


bool View::usesPref(const QString &s) {
   return (s.left(9) == "key.tool." || s.left(10) == "stacking.")
    ? true : false;
}


QValueList<Token *> View::tokens() const {
   QValueList<Token *> list;
   QCanvasItemList l = removeNonTokens(canvas()->allItems());
   for (QCanvasItemList::const_iterator i = l.begin(); i != l.end(); ++i)
    list.append((Token *)*i);
   return list;
}


QCanvasItemList View::removeNonTokens(QCanvasItemList l) const {
   for (QCanvasItemList::iterator i = l.begin(); i != l.end(); ++i)
    if ((*i)->rtti() != Token::RTTI) { i = l.remove(i); --i; }
  
   return l;
}


void View::rebindKey(const QString binding) {
   // ignore if not a key binding
   if (binding.left(4) != "key.") return;

   int oldkey = binding2key[binding];
   int newkey = Pref[binding].effective().toInt();
   binding2key[binding] = newkey;

   // disconnect old binding
   accel.removeItem(oldkey);

   // connect new binding
   accel.insertItem(newkey, newkey);
   accel.connectItem(newkey, MW, Slot[binding]);
}


void View::keyPressEvent(QKeyEvent *e) {
   if (MW->tool() == TSelect) {
      if (e->key() == Key_Shift) {
         if (e->state() & ControlButton) setCursor(*IntersectCursor);
         else setCursor(*UnionCursor);
         e->accept();
      }
      else if (e->key() == Key_Control) {        
         if (e->state() & ShiftButton) setCursor(*IntersectCursor);
         else setCursor(*MinusCursor);
         e->accept();
      }          
   }
   else if (e->key() == Key_Shift) {
      setCursor(*PlusCursor);
      e->accept();
   }     
   else e->ignore();
}


void View::keyReleaseEvent(QKeyEvent *e) {
   if (MW->tool() == TSelect) {
      if (e->key() == Key_Shift) {
         if (e->state() & ControlButton) setCursor(*MinusCursor);
         else setCursor(*NormalCursor);
         e->accept();
      }
      else if (e->key() == Key_Control) {        
         if (e->state() & ShiftButton) setCursor(*UnionCursor);
         else setCursor(*NormalCursor);
         e->accept();
      }        
   }
   else if (e->key() == Key_Shift) {
      setCursor(*NormalCursor);
      e->accept();
   }
   else e->ignore();   
}


void View::contentsMousePressEvent(QMouseEvent *e) {
   // find collisions
   collisions = removeNonTokens(canvas()->collisions(e->pos()));   
   if ((e->state() & ShiftButton) && collisions.begin() != collisions.end()) {
      collisions = collisionsRecursive(collisions.first(), collisions);
      collisions.sort();
   }

   // make a local copy of the event so we can change it
   QMouseEvent *event = new QMouseEvent(QEvent::MouseButtonPress,
    e->pos(), e->button(), e->state());
   
   switch (event->button()) {
      case LeftButton:
         if (MW->tool() == TSelect || MW->tool() == TWand) {
            // do nothing
         }
         else if (game->selector->count()) {
            // Check whether anything in the selection was clicked.
            bool clear = true;

            QCanvasItemList s = game->selector->selection();
            for (QCanvasItemList::const_iterator i = s.begin();
             i != s.end() && clear; ++i) {
               if (collisions.find(*i) != collisions.end()) clear = false;
            }

            // If not, clear selection. Otherwise collisions = selected.
            if (clear) game->selector->clearSelection();
            else {
               collisions = s;
               delete event;
               event = new QMouseEvent(QEvent::MouseButtonPress,
               e->pos(), e->button(), e->state() | ShiftButton);
            }            
         }
         
         // Is control depressed? If not, apply current tool.   
         if ((event->state() & ControlButton) && MW->tool() != TSelect &&
          MW->tool() != TCrayon) { 
            // start a new compound action
            // FIXME: reinstate once TokenBeginCompoundAction is converted
            /*
            TokenBeginCompoundAction *a = new TokenBeginCompoundAction;
            a->commit();
            Game->done.log(a);
            compound = true;
            */
         }
         else (this->*toolFunc[MW->tool()])(event);

         break;    
      
      case RightButton:
         if (plot) {
            // FIXME: reinstate once TokenEndPlotAction is converted
            /*
            TokenEndPlotAction *epa =
             new TokenEndPlotAction(this, (FreehandItem*)plot, plotting);
            QPointArray pa = ((FreehandItem*)plot)->points();
            QPoint d = pa[0] - pa[int(pa.size()-1)];
            
            for (QCanvasItemList::const_iterator i = plotting.begin();
             i != plotting.end(); ++i) {
               TokenMoveAction *a = new TokenMoveAction((Token*)*i,
                (*i)->x() - d.x(), (*i)->y() - d.y(), ++max_height);
               epa->append(a);
            }
            
            epa->commit();
            Game->done.log(epa);
            
            plot = 0;
            plotting.clear();
            */
         }
         else if (compound) {
            // FIXME: reinstate once TokenEndCompoundAction is converted
            /*
            // close compound action
            TokenEndCompoundAction *a = new TokenEndCompoundAction;
            a->commit();
            Game->done.log(a);
            compound = false;
            */
         }
         else {
            (this->*toolFunc[TRaise])(event);
             // if no tokens to raise, then bring up menu
            if (!raised) menu.popup(QCursor::pos());
         }
         break;
         
      case MidButton:
      default:
         break;
   }

   delete event;
}


void View::contentsMouseReleaseEvent(QMouseEvent *e) {
   if (moving.count()) endMovement();
   else if (raised) {
      raised->setZ(fall);
      canvas()->update();
      raised = 0;
   }
   else if (selecting) {
      // kill the selecting rectangle
      QRect r = ((SelectingItem*)selecting)->rect().normalize();
      delete selecting;
      selecting = 0;
      
      QCanvasItemList l = removeNonTokens(canvas()->collisions(r));

      // intersect
      if ((e->state() & ShiftButton) && (e->state() & ControlButton))
       game->selector->intersectSelection(l);
      // subtract
      else if (e->state() & ControlButton)
       game->selector->subtractSelection(l);
      // union
      else if (e->state() & ShiftButton)
       game->selector->unionSelection(l);
      // assign
      else if (l.count())
       game->selector->assignSelection(l);
      // clear
      else if (game->selector->count())
       game->selector->clearSelection();
      // no operation, clear the selection rectangle
      else canvas()->update();
   }
   
   if (freehand) {
      // stop drawing
      //freehand = 0;

      // FIXME: do something to keep track of FreehandItems
   }
}


void View::contentsMouseMoveEvent(QMouseEvent *e) {
   if (moving.count()) {
     if (!scroll_timer.isActive()) {
         QPoint d = e->pos() - moving_start;
         bool drag, scroll;
      
         clipMovement(d, &drag, &scroll);
      
         QCanvasItemList draglist;
         QValueList<QPoint> points;
        
         // FIXME: scrolling a token to an edge should start a drag

         // move the tokens 
         for (QCanvasItemList::const_iterator i = moving.begin();
          i != moving.end(); ++i) {
            if ((*i)->rtti() != Token::RTTI) continue;
            
            (*i)->moveBy(d.x(), d.y());
            
            if (drag) {
               draglist.append(*i);
               points.append(QPoint(e->pos().x() - (int)(*i)->x(),
                e->pos().y() - (int)(*i)->y()));
            }
         }         

         if (drag) {
            // start DnD at edges
            stopAutoScroll();
            endMovement();           
 
            QDragObject *d = new TokenDrag(draglist, points, this);
            if (MW->tool() == TMove) d->dragMove();
            else d->dragCopy();       
         }
         else if (scroll) startAutoScroll();

         moving_start = e->pos();
         canvas()->update();
     }
   }        
   else if (selecting) {
      // FIXME: try resizing the rectangle rather than deleting
      QPoint o = ((SelectingItem*)selecting)->rect().topLeft();
      delete selecting;
    
      // prevent selection box from exceeding canvas bounds 
      int x = QMAX(QMIN(e->pos().x(), canvas()->width()-1), 0);
      int y = QMAX(QMIN(e->pos().y(), canvas()->height()-1), 0);
      
      // draw new selection box
      selecting = new SelectingItem(QRect(o, QPoint(x, y)), canvas());
      selecting->setZ(++max_height);
      selecting->show();
      canvas()->update();
   }

   /*              
   if (freehand) {
      //((FreehandItem*)freehand)->appendPoint(e->pos());
      ((CrayonMark *)freehand)->pushPoint(e->pos());
      canvas()->update();      
   }
   */
}


void View::contentsDragEnterEvent(QDragEnterEvent *e) {
   e->accept(TokenDrag::canDecode(e));       
}


void View::contentsDropEvent(QDropEvent *e) {
   QCanvasItemList list;
   QValueList<QPoint> points;
   
   if (TokenDrag::decode(e, list, points)) {
      QCanvas *cn = list.first()->canvas();

      QValueList<QPoint>::const_iterator pi = points.begin();
      for (QCanvasItemList::const_iterator ci = list.begin(); ci != list.end();
       ++ci, ++pi) {
         Token *c = (Token*)*ci;
         QPoint p = *pi;
                 
         c->setCanvas(canvas());
  
         int x = QMAX(QMIN(e->pos().x() - p.x(),
          canvas()->width() - c->boundingRect().width()), 0);
         int y = QMAX(QMIN(e->pos().y() - p.y(),
          canvas()->height() - c->boundingRect().height()), 0);
          
         c->setX(x);
         c->setY(y);
         c->setZ(++max_height);

         if (dynamic_cast<TokenCreateAction *>(game->done.top())) {
            game->inplaydialog->insert(c);
            game->groupsdialog->insert(c);         
         }
      }
         
      cn->update();
      canvas()->update();
   }

   endMovement();
}


void View::setPlotting(FreehandItem *p, QCanvasItemList l) {
   plot = p;
   plotting = l;
}


void View::clipMovement(QPoint &p, bool *drag, bool *scroll) {
   // find bounding rectangle
   QRect b;
   for (QCanvasItemList::const_iterator i = moving.begin(); i != moving.end();
    i++) if ((*i)->rtti() == Token::RTTI) b |= (*i)->boundingRect();

   // get visible rectangle
   QRect v(contentsX(), contentsY(), visibleWidth(), visibleHeight());
   
   // start scrolling?
   if (scroll) *scroll = !v.contains(b) ? true : false; 
   
   // adjust movement to prevent overrunning canvas edges
   if (drag) *drag = false;
        
   if (b.left() + p.x() < 0) {
      p.setX(-b.left());
      if (drag) *drag = true;
   }
   else if (b.right() + p.x() > canvas()->width()) {
      p.setX(canvas()->width()-b.right());
      if (drag) *drag = true;
   }
   
   if (b.top() + p.y() < 0) {
      p.setY(-b.top());
      if (drag) *drag = true;
   }
   else if (b.bottom() + p.y() > canvas()->height()) {
      p.setY(canvas()->height()-b.bottom());
      if (drag) *drag = true;
   }
}


void View::endMovement() {
   // signal that movement is finished   
   game->done.complete();
   moving.clear();
}
        

void View::cut() { cut(game->selector->selection()); }


void View::cut(QCanvasItemList list) {
   if (list.isEmpty()) return;
   
   ComboAction *tcut = new ComboAction;

   // gather tokens 
   QValueList<QPoint> points;
   QPoint min(canvas()->rect().bottomRight());

   for (QCanvasItemList::iterator i = list.begin();
    i != list.end(); ++i) {
      QCanvasItem *s = *i;
      if (s->rtti() != Token::RTTI) { list.remove(s); continue; }

      Token *t = (Token *)s;
      TokenTrashAction *a = new TokenTrashAction(t);

      if (a->isPermissible()) {
         tcut->append(a);
        
         QPoint p((int)t->x(), (int)t->y());
         points.append(p);

         if (min.x() > p.x()) min.setX(p.x());
         if (min.y() > p.y()) min.setY(p.y());

         Token *c = new Token(*t, game->maxID++);
         game->gid2g[c->id()] = c;

         c->reparent(0);
         c->setCanvas(game->scratch);

         *i = c;
      }
      else {
         list.remove(*i);
         --i;
         delete a;
      }
   }

   if (tcut->isPermissible()) {
      game->selector->clearSelection();
      clearClipboard();

      // normalize points to minimum point
      for (QValueList<QPoint>::iterator p = points.begin();
       p != points.end(); ++p) *p -= min;

      TokenDrag *d = new TokenDrag(list, points);
      QApplication::clipboard()->setData(d);

      logAction(tcut);
   }
   else delete tcut;

   // enable paste
   setPasteable(true);
}


void View::copy() { copy(game->selector->selection()); }


void View::copy(QCanvasItemList list) {
   if (list.isEmpty()) return;

   // gather tokens 
   QValueList<QPoint> points;
   QPoint min(canvas()->rect().bottomRight());

   for (QCanvasItemList::iterator i = list.begin(); i != list.end(); ++i) {
      QCanvasItem *s = *i;
      if (s->rtti() != Token::RTTI) { list.remove(s); continue; }

      // FIXME: This is weird; we don't actually use the TokenCloneAction.
      Token *t = (Token *)s;
      TokenCloneAction *a = new TokenCloneAction(t, 0);

      if (a->isPermissible()) {
         QPoint p((int)t->x(), (int)t->y());
         points.append(p);

         if (min.x() > p.x()) min.setX(p.x());
         if (min.y() > p.y()) min.setY(p.y());

         Token *c = new Token(*t, game->maxID++);
         game->gid2g[c->id()] = c;

         c->reparent(0);
         c->setCanvas(game->scratch);

         *i = c;
      }
      else {
         list.remove(*i);
         --i;
      }

      delete a;
   }

   if (list.count()) {
      // clear the clipboard
      clearClipboard();

      // normalize points to minimum point
      for (QValueList<QPoint>::iterator p = points.begin();
       p != points.end(); ++p) *p -= min;

      // put the token information on the clipboard
      TokenDrag *d = new TokenDrag(list, points);
      QApplication::clipboard()->setData(d);
   }

   // enable paste
   setPasteable(true);
}


void View::paste() {
   if (!TokenDrag::canDecode(QApplication::clipboard()->data())) return;

   QCanvasItemList list;
   QValueList<QPoint> points;  

   if (!TokenDrag::decode(QApplication::clipboard()->data(), list, points))
    return;

   ComboAction *tpaste = new ComboAction;

   // create the individual TokenCreateActions
   QValueList<QPoint>::const_iterator pi = points.fromLast();
   for (QCanvasItemList::const_iterator ci = list.fromLast();
    ci != list.end(); --ci, --pi) {
      Token *t = (Token *)*ci;
      int x = QMAX(QMIN(QCursor::pos().x() - geometry().x() + (*pi).x() +
       contentsX(), canvas()->width() - t->boundingRect().width()), 0);
      int y = QMAX(QMIN(QCursor::pos().y() - geometry().y() + (*pi).y() +
       contentsY(), canvas()->height() - t->boundingRect().height()), 0);

      TokenCloneAction *a = new TokenCloneAction(t,
       game->groupsdialog->currentGroup(), canvas(), x, y, ++max_height);

      if (a->isPermissible()) tpaste->append(a);
      else {
         delete a;
         list.remove(*ci);
         --ci;
      }
   }

   if (!tpaste->count()) delete tpaste;
   else { 
      logAction(tpaste);
      game->selector->assignSelection(tpaste->createdList());
   }
}


void View::clearClipboard() {
   // clean up scratch if we can read what's on the clipboard
   if (!TokenDrag::canDecode(QApplication::clipboard()->data())) return;

   QCanvasItemList l;
   QValueList<QPoint> points;  
 
   if (!TokenDrag::decode(QApplication::clipboard()->data(), l, points))
    return;

   for (QCanvasItemList::const_iterator i = l.begin(); i != l.end(); ++i) {
      if ((*i)->canvas() == game->scratch) {
         if ((*i)->rtti() == Token::RTTI) {
            Token *t = (Token *)*i;
            game->gid2g.remove(t->id());
         }

         delete *i;
      }
   }
 
   QApplication::clipboard()->clear(); 
}


void View::normalizeTokenHeights() {
   QCanvasItemList l = removeNonTokens(canvas()->allItems());
   l.sort();
 
   // normalize token heights
   unsigned int j = l.count();
   max_height = j;
   for (QCanvasItemList::const_iterator i = l.begin(); i != l.end(); ++i)
    (*i)->setZ(j--);
}


void View::setSelected(bool enable) {
   menu.setItemEnabled(mCut, enable);
   menu.setItemEnabled(mCopy, enable);
   menu.setItemEnabled(mStackOffset, enable);
   menu.setItemEnabled(mStackFlush, enable);
   menu.setItemEnabled(mStackShuffle, enable);
   menu.setItemEnabled(mStackInvert, enable);
}


void View::setPasteable(bool enable) {
   menu.setItemEnabled(mPaste, enable);
}


QCanvasItemList View::collisionsRecursive(const QCanvasItem *c,
 QCanvasItemList r) const {
   QCanvasItemList l = removeNonTokens(c->collisions(true)); 
   for (QCanvasItemList::const_iterator i = l.begin(); i != l.end(); ++i) {
      if (r.find(*i) == r.end()) {
         r.append(*i);
         r = collisionsRecursive(*i, r);
      }
   }
   return r;
}


void View::scrollUp() { scrollBy(0, -10); }
void View::scrollDown() { scrollBy(0, 10); }
void View::scrollLeft() { scrollBy(-10, 0); }
void View::scrollRight() { scrollBy(10, 0); }


void View::startAutoScroll() {
   if (!scroll_timer.isActive()) {
      scroll_timer.start(30);
      scroll_move = 10;
   }
}


void View::doAutoScroll() {
   QPoint v = viewport()->mapFromGlobal(QCursor::pos());
   QPoint p = viewportToContents(v);

   scroll_timer.start(30);
   // FIXME: allow adjustable scroll speed
   scroll_move=8*log(scroll_move);
   
   QPoint d(0,0);
   
   if (v.x() < ScrollMargin) d.setX((int)-scroll_move);
   else if (v.x() > visibleWidth() - ScrollMargin) d.setX((int)+scroll_move);
   
   if (v.y() < ScrollMargin) d.setY((int)-scroll_move);
   else if (v.y() > visibleHeight() - ScrollMargin) d.setY((int)+scroll_move);
 
   if (!d.isNull()) {
      scrollBy(d.x(),d.y());  
     
      // prevent edge overruns      
      QPoint m = p - moving_start;
      clipMovement(m); 
      
      for (QCanvasItemList::const_iterator i = moving.begin();
       i != moving.end(); ++i) (*i)->moveBy(m.x(), m.y());
      moving_start = p;
      canvas()->update();
   }   
   else stopAutoScroll();
}
        

void View::stopAutoScroll() { scroll_timer.stop(); }


double View::zoom() const { return Zoom; }


/*
void View::drawContents(QPainter *p, int cx, int cy, int cw, int ch) {
   if (Zoom != 1) p->scale(Zoom, Zoom);
   QCanvasView::drawContents(p, cx, cy, cw, ch);
}
*/        


void View::logAction(Action *a) {  
   a->commit();
   game->done.log(a);
   game->done.complete();
}


void View::selectAll() {
   game->selector->assignSelection(removeNonTokens(canvas()->allItems()));
}


void View::selectNone() { game->selector->clearSelection(); }
        

void View::selectInvert() {
   QCanvasItemList all = removeNonTokens(canvas()->allItems());
   QCanvasItemList inv;
   QCanvasItemList selected = game->selector->selection();

   for (QCanvasItemList::const_iterator i = all.begin(); i != all.end(); ++i)
    if (selected.find(*i) == selected.end()) inv.append(*i);

   game->selector->assignSelection(inv);
}
        

void View::stackOffset() {
   stack(game->selector->selection(), QPoint(
    Pref["stacking.offset.x"].effective().toInt(),
    Pref["stacking.offset.y"].effective().toInt()));
}        
        

void View::stackFlush() { stack(game->selector->selection(), QPoint(0,0)); }


void View::stack(QCanvasItemList l, const QPoint &offset) {
   int i = l.count() - 1;
   double x = l[i]->x() + offset.x();
   double y = l[i]->y() + offset.y();

   TokenStackAction *s = new TokenStackAction;   
   for (i--; i >= 0; i--, x += offset.x(), y += offset.y()) {   
      if (l[i]->rtti() == Token::RTTI) {
         Token *c = (Token *)l[i];
         TokenMoveAction *a = new TokenMoveAction(c, x, y, c->z());
         if (a->isPermissible()) s->append(a);
         else delete a;   
      }
   }

   if (s->count()) logAction(s);
   else delete s;
}


void View::stackShuffle() { stackShuffle(game->selector->selection()); }


void View::stackShuffle(QCanvasItemList l) {
   const unsigned int count = l.count();
   //unsigned int n[count];
   unsigned int *n = new unsigned int[count];

   // fill shuffle array
   for (unsigned int i = 0; i < count; i++) n[i] = i;

   // shuffle shuffle array
   for (unsigned int i = 0; i < count; i++) {
      unsigned int r = (unsigned int)(((double)count*rand())/(RAND_MAX+1.0));
      unsigned int t = n[r];
      n[r] = n[i];
      n[i] = t;
   }

   // reposition tokens according to shuffle array
   TokenShuffleAction *s = new TokenShuffleAction;   
   for (unsigned int i = 0; i < count; i++) {
      if (l[n[i]]->rtti() == Token::RTTI) {
         Token *c = (Token *)l[n[i]];
         TokenMoveAction *a = new
          TokenMoveAction(c, l[i]->x(), l[i]->y(), l[i]->z());
         if (a->isPermissible()) s->append(a);
         else delete a;
      } 
   }  
   
   delete[] n;

   if (s->count()) logAction(s);
   else delete s;
}


void View::stackInvert() { stackInvert(game->selector->selection()); }


void View::stackInvert(QCanvasItemList l) {
   TokenInvertAction *s = new TokenInvertAction;   
   for (unsigned int i = 0, j = l.count() - 1; i <= j; i++, j--) {   
      if (l[i]->rtti() == Token::RTTI && l[j]->rtti() == Token::RTTI) {
         Token *c1 = (Token *)l[i];
         Token *c2 = (Token *)l[j];
         
         TokenMoveAction *a1 =
          new TokenMoveAction(c1, c1->x(), c1->y(), c2->z());
         TokenMoveAction *a2 =
          new TokenMoveAction(c2, c2->x(), c2->y(), c1->z());
         
         if (a1->isPermissible() && a2->isPermissible()) {
            s->append(a1);
            s->append(a2);
         }
         else {
            delete a1;
            delete a2;
         }
      }
   }

   if (s->count()) logAction(s);
   else delete s;

   if (Pref["stacking.restack_after_invert"].effective().toBool())
    stackOffset();
}


void View::properties() {
   ViewPropertiesDialog *propDialog = new ViewPropertiesDialog;
   propDialog->show();
}


void View::toolMove(QMouseEvent *e) {
   moving_start = e->pos();
   if (e->state() & ShiftButton) {
      ComboAction *combo = new ComboAction;
      // iterate from bottom to top when Shift depressed
      for (QCanvasItemList::const_iterator i = collisions.fromLast();
       i != collisions.end(); --i) {
         if ((*i)->rtti() == Token::RTTI) {
            Token *c = (Token *)*i;
            TokenAction *a = new TokenMoveAction(c);
            if (a->isPermissible()) {
               ((TokenMoveAction *)a)->begin();
               combo->append(a);
               moving.append(c);
               c->setZ(++max_height);
            }
            else delete a;
         }
      }
   
      Action *a = combo->collapse();
      if (a) game->done.log(a);
   }
   else {
      // iterate from top to bottom
      for (QCanvasItemList::const_iterator i = collisions.begin();
       i != collisions.end(); ++i) {
         if ((*i)->rtti() == Token::RTTI) {
            Token *c = (Token *)*i;
            Action *a = new TokenMoveAction(c);
            if (a->isPermissible()) {
               ((TokenMoveAction *)a)->begin();
               moving.append(c);
               c->setZ(++max_height);
               game->done.log(a);
               canvas()->update();
            }
            else delete a;
            break;
         }
      }
   }
}


void View::toolClone(QMouseEvent *e) {
   moving_start = e->pos();
   if (e->state() & ShiftButton) {
      ComboAction *combo = new ComboAction;
      // iterate from bottom to top when Shift depressed
      for (QCanvasItemList::const_iterator i = collisions.fromLast();
       i != collisions.end(); --i) {
         if ((*i)->rtti() == Token::RTTI) {
            Token *c = (Token *)*i;
            TokenAction *a = new TokenCloneAction(c,
             game->groupsdialog->currentGroup());
            if (a->isPermissible()) {
               a->begin();
               c = ((TokenCloneAction *)a)->created();
               combo->append(a);
               moving.append(c);
               c->setZ(++max_height);
            }
            else delete a;
         }
      }
     
      Action *a = combo->collapse();
      if (a) game->done.log(a);
   }
   else {
      // iterate from top to bottom
      for (QCanvasItemList::const_iterator i = collisions.begin();
       i != collisions.end(); ++i) {
         if ((*i)->rtti() == Token::RTTI) {
            Token *c = (Token *)*i;
            Action *a = new TokenCloneAction(c,
             game->groupsdialog->currentGroup());
            if (a->isPermissible()) {
               a->begin();
               c = ((TokenCloneAction *)a)->created();
               moving.append(c);
               c->setZ(++max_height);
               game->done.log(a);
               canvas()->update();
            }
            else delete a;
            break;
         }
      }
   }
}


void View::toolPlot(QMouseEvent *e) {
   // FIXME: reinstate once plot actions are converted
   /*
   if (plot) {
      // add a new plot point
      TokenPointPlotAction *a =
       new TokenPointPlotAction(this, (FreehandItem*)plot,
       plotting, e->pos());
      if (a->isPermissible()) {
         a->begin();
         game->done.log(a);
      }
      else delete a;
   }
   else {
      // start new plotted move
           
      // FIXME: Won't draw first segment without two points to
      // start. Why?
      QPointArray p(2);
      p[0] = p[1] = e->pos();
      plot = new FreehandItem(p, canvas());
      ((FreehandItem*)plot)->setPen(QPen("green", 4,
       Qt::SolidLine));
      plot->show();
      
      TokenBeginPlotAction *a =
       new TokenBeginPlotAction(this, (FreehandItem*)plot, plotting);

      if (e->state() & ShiftButton) {
         // iterate from bottom to top
         for (QCanvasItemList::const_iterator i = collisions.fromLast();
          i != collisions.end(); --i) {
            if ((*i)->rtti() == Token::RTTI) {
               Token *c = (Token *)*i;
               Action *m = new TokenMoveAction(c);
               // add token to plot only if a move is allowed
               if (m->isPermissible()) plotting.append(*i);
               delete m;
            }
         }
      }
      else {
         // iterate from top to bottom
         for (QCanvasItemList::const_iterator i = collisions.begin();
          i != collisions.end(); ++i) { 
            if ((*i)->rtti() == Token::RTTI) {
               Token *c = (Token *)*i;
               Action *m = new TokenMoveAction(c);
               // add token to plot only if a move is allowed
               if (m->isPermissible()) plotting.append(*i);
               delete m;
               break;
            }
         } 
      }  
      
      // start plot only if some tokens were included
      if (plotting.count()) {
         a->begin();
         Game->done.log(a);
      }
      else {
         delete plot;
         delete a;
         plot = 0;
      }
   }
   */
}


void View::toolTrash(QMouseEvent *e) {
   ComboAction *combo = new ComboAction;
   Action *a;
   for (QCanvasItemList::const_iterator i = collisions.begin();
    i != collisions.end(); ++i) {
      if ((*i)->rtti() == Token::RTTI) {
         Token *c = (Token*)*i;
         a = new TokenTrashAction(c);
         if (a->isPermissible()) combo->append((TokenAction *)a);
         else delete a;
         if (!(e->state() & ShiftButton)) break;
      }               
   } 
 
   a = combo->collapse();
   if (a) logAction(a);
}


void View::toolFlipLeft(QMouseEvent *e) {
   ComboAction *combo = new ComboAction;
   Action *a;
   for (QCanvasItemList::const_iterator i = collisions.begin();
    i != collisions.end(); ++i) {
      if ((*i)->rtti() == Token::RTTI) {
         Token *c = (Token *)*i;
         a = new TokenFlipLeftAction(c);
         if (a->isPermissible()) combo->append((TokenAction *)a);
         else delete a;
         if (!(e->state() & ShiftButton)) break;
      }               
   } 
  
   a = combo->collapse();
   if (a) logAction(a); 
}


void View::toolFlipRight(QMouseEvent *e) {
   ComboAction *combo = new ComboAction;
   Action *a;
   for (QCanvasItemList::const_iterator i = collisions.begin();
    i != collisions.end(); ++i) {
      if ((*i)->rtti() == Token::RTTI) {
         Token *c = (Token *)*i;
         a = new TokenFlipRightAction(c);
         if (a->isPermissible()) combo->append((TokenAction *)a);
         else delete a;
         if (!(e->state() & ShiftButton)) break;
      }               
   } 
  
   a = combo->collapse();
   if (a) logAction(a);
}


void View::toolRotateCW(QMouseEvent *e) { 
   ComboAction *combo = new ComboAction;
   Action *a;
   for (QCanvasItemList::const_iterator i = collisions.begin();
    i != collisions.end(); ++i) {
      if ((*i)->rtti() == Token::RTTI) {
         Token *c = (Token *)*i;
         a = new TokenRotateCWAction(c);
         if (a->isPermissible()) combo->append((TokenAction *)a);
         else delete a;
         if (!(e->state() & ShiftButton)) break;
      }               
   } 
   
   a = combo->collapse();
   if (a) logAction(a);
}


void View::toolRotateCCW(QMouseEvent *e) {
   ComboAction *combo = new ComboAction;
   Action *a;
   for (QCanvasItemList::const_iterator i = collisions.begin();
    i != collisions.end(); ++i) {
      if ((*i)->rtti() == Token::RTTI) {
         Token *c = (Token *)*i;
         a = new TokenRotateCCWAction(c);
         if (a->isPermissible()) combo->append((TokenAction *)a);
         else delete a;
         if (!(e->state() & ShiftButton)) break;
      }               
   } 
   
   a = combo->collapse();
   if (a) logAction(a);
}


void View::toolRegroup(QMouseEvent *e) {
   Group *group = game->groupsdialog->currentGroup();
   ComboAction *combo = new ComboAction;
   Action *a;   
   for (QCanvasItemList::const_iterator i = collisions.begin();
    i != collisions.end(); ++i) {
      if ((*i)->rtti() == Token::RTTI) {
         Token *c = (Token *)*i;
         a = new TokenRegroupAction(c, group);
         if (a->isPermissible()) combo->append((TokenAction *)a);
         else delete a;
         if (!(e->state() & ShiftButton)) break;
      }
   } 
   
   a = combo->collapse();
   if (a) logAction(a);
}


/*
void View::toolGroup(QMouseEvent *e) {
   GroupListBoxItem *group = MW->group();
   ComboAction *combo = new ComboAction;
   Action *a;
   for (QCanvasItemList::Iterator i = collisions.begin();
    i != collisions.end(); i++) {
      if ((*i)->rtti() == Token::RTTI) {
         Token *c = (Token *)*i;
         a = new TokenGroupAction(c, group);
         if (a->isPermissible()) combo->append((TokenAction *)a);
         else delete a;
         if (!(e->state() & ShiftButton)) break;
      }               
   } 
  
   if ((a = collapseCombo(combo))) {
      a->commit();
      game->done.log(a);
   }
}


void View::toolUngroup(QMouseEvent *e) {
   GroupListBoxItem *group = MW->group();
   ComboAction *combo = new ComboAction;
   Action *a;
   for (QCanvasItemList::Iterator i = collisions.begin();
    i != collisions.end(); i++) {
      if ((*i)->rtti() == Token::RTTI) {
         Token *c = (Token *)*i;
         a = new TokenUngroupAction(c, group);
         if (a->isPermissible()) combo->append((TokenAction *)a);
         else delete a;
         if (!(e->state() & ShiftButton)) break;
      }               
   } 
  
   if ((a = collapseCombo(combo))) {
      a->commit();
      game->done.log(a);
   }       
}
*/


void View::toolSelect(QMouseEvent *e) {
   selecting = new SelectingItem(QRect(e->pos(), e->pos()), canvas());
}


void View::toolWand(QMouseEvent *e) {
   if (collisions.count()) { 
      QCanvasItemList l;   
      if (e->state() & ShiftButton) l = collisions;
      else l.append(collisions.first());
     
      game->selector->unionSelection(l);
   }
   else if (game->selector->count()) game->selector->clearSelection();
}
        

void View::toolCrayon(QMouseEvent *e) {
   //freehand = new FreehandItem(p, canvas());
   //((FreehandItem*)freehand)->setPen(QPen("red", 2, Qt::SolidLine));
   
   /*
   if (scrawl) {
      ((CrayonMark *)freehand)->pushPoint(e->pos());
   }
   else { 
      freehand = new CrayonMark(e->pos(), canvas());
      freehand->show();
   }
   */

   QPen pen(game->user->drawingColor(game->user),
    game->user->drawingWidth(game->user));

   if (!pline) {
      PolylineBeginAction *pb = new PolylineBeginAction(game->user, canvas(),
       e->pos().x(), e->pos().y(), "foo", pen);

      if (pb->isPermissible()) {
         logAction(pb);
         pline = pb->created();
      }
      else delete pb;
   }
   else {
      PolylineExtendAction *pe =
       new PolylineExtendAction(pline, e->pos().x(), e->pos().y(), pen);

      if (pe->isPermissible()) {
         logAction(pe);
         if ((e->state() & ControlButton)) pline = 0;
      }
      else delete pe;
   }
}


void View::toolZoomIn(QMouseEvent *) {
}


void View::toolZoomOut(QMouseEvent *) {
}


void View::toolProp(QMouseEvent *) {
   /*
   for (QCanvasItemList::Iterator i = collisions.begin();
    i != collisions.end(); i++) {
      if ((*i)->rtti() == Token::RTTI) {
         Token *c = (Token *)*i;
         TokenPropertiesDialog *chitDialog = new TokenPropertiesDialog(c);
         chitDialog->show();
         break;
      }               
   }
   */ 
}
        

void View::toolRaise(QMouseEvent *) {
   for (QCanvasItemList::const_iterator i = collisions.begin();
    i != collisions.end(); ++i) {
      if ((*i)->rtti() == Token::RTTI) {
         raised = *i; 
         break;
      }
   }

   if (raised) {
      fall = raised->z();
      raised->setZ(++max_height);
      canvas()->update();
   }            
}
